#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/06/2020
           """

__all__ = [
    "is_module_available",
    "import_warning",
    "reimported_warning",
    "ensure_in_sys_path",
    "clean_sys_path",
    "remove_from_sys_path",
    "import_file",
    "find_ancestral_relatives",
    "find_nearest_ancestral_relative",
    "walk_up",
    "reload_all_modules",
    "reload_module",
]

import importlib
import sys
from importlib import reload
from importlib.util import find_spec
from pathlib import Path
from typing import Optional, Any, Union, List, Iterable
from warnings import warn

import pkg_resources

from warg import passes_kws_to

"""
PRELOADED_MODULES = set()
def init():
  # local imports to keep things neat
  from sys import modules

  global PRELOADED_MODULES

  # sys and importlib are ignored here too
  PRELOADED_MODULES = set(modules.values())


def reload_all():
  from sys import modules
  import importlib

  for module in set(modules.values()) - PRELOADED_MODULES:
    try:
      importlib.reload(module)
    except:
      # there are some problems that are swept under the rug here
      pass
init()
"""


def contain(q: Any, s: Iterable) -> bool:
    return q in s


def reload_module(module_name: str, containment_test: callable = contain) -> None:
    if module_name in sys.modules:
        reload_set = {x for x in sys.modules if containment_test(module_name, x)}
        for a in reload_set:
            del sys.modules[a]
            # importlib.reload(sys.modules[mod_str]) #DOES NOT WORK ON FROM IMPORTS...
            sys.modules[a] = importlib.import_module(a)
    else:
        sys.modules[module_name] = importlib.import_module(module_name)


def reload_requirements(requirements_path: Path, containment_test: callable = contain) -> None:
    with open(requirements_path) as f:
        for r in pkg_resources.parse_requirements(f.readlines()):
            reload_module(r.project_name, containment_test=containment_test)


def reload_all_modules(catch_exceptions: bool = True, verbose: bool = True) -> None:
    try:
        for mod in sys.modules.values():
            reload(mod)
    except Exception as e:
        if verbose:
            print(mod)
        if catch_exceptions:
            if verbose:
                print(e)
        else:
            raise e


def import_file(path: Path, from_list=None) -> Any:
    """Import a module given its filename, works both on absolute and relative paths"""
    if from_list is None:
        from_list = {}
    globals_ = {}  # globals() # determines package context
    locals_ = {}  # locals() # Should not be used in import anyway

    sys_path = sys.path  # Save original sys.path
    try:
        sys.path.insert(0, str(path.parent.absolute()))  # Temporarily add parent dir of path to parent
        return __import__(
            path.stem, globals=globals_, locals=locals_, fromlist=from_list, level=0
        )  # Get the module name (no extension)
    finally:
        sys.path = sys_path  # Restore original sys.path


def walk_up(path: Path, top: Path, max_ascent: int = None):
    i = 0
    while True:
        yield path
        i += 1
        if max_ascent and max_ascent < i:
            break
        if path == top:
            break
        else:
            path = path.parent


def walk_down(path: Path, max_descent: int = None):
    if max_descent == 0:
        return

    queue = []
    for c in path.iterdir():
        if c.is_dir():
            yield c
            queue.append(c)

    for q in queue:
        try:
            yield from walk_down(q, max_descent=max_descent - 1 if max_descent else None)
        except:
            yield


def find_ancestral_relatives(
    target: Union[str, Path],
    context: Path = Path.cwd(),
    *,
    from_parent_of_context: bool = True,
    ancestral_levels: int = 2,
    descendant_levels: int = 2,
    top_level: Path = None,
    return_parent_of_target: bool = True,
    no_duplicates: bool = True,
    terminate_first: bool = False,
) -> List[Path]:
    relatives = []

    if top_level is None:
        top_level = context.root

    if from_parent_of_context:
        context = context.parent

    for p in walk_up(context, top_level, max_ascent=ancestral_levels):
        for wd in walk_down(p, max_descent=descendant_levels):
            if target in wd.parts[-ancestral_levels:]:
                if return_parent_of_target:
                    wd = wd.parent
                relatives.append(wd)
                if terminate_first:
                    break
        p /= target
        if p.exists():
            if return_parent_of_target:
                p = p.parent
            relatives.append(p)
            if terminate_first:
                break

    if no_duplicates:
        return list(set(relatives))

    return relatives


@passes_kws_to(find_ancestral_relatives)
def find_nearest_ancestral_relative(*args, **kwargs) -> Optional[Path]:
    kwargs.update(terminate_first=True)
    result = find_ancestral_relatives(*args, **kwargs)
    if result:
        return result[0]


def clean_sys_path() -> None:
    """
    Clean the sys.path for dead paths or duplicates
    """
    out = []
    for path in sys.path:
        p = Path(path).resolve()
        if p.exists():
            if p not in out:
                out.append(p)

    sys.path[:] = [str(o.absolute()) for o in out]


def remove_from_sys_path(target: Path, missing_ok: bool = True):
    """
    Clean the sys.path for dead paths or duplicates
    """
    out = []

    target = target.resolve()

    for path in sys.path:
        p = Path(path).resolve()
        if p.exists() and target != p:
            if p not in out:
                out.append(p)

    sys.path[:] = [str(o.absolute()) for o in out]


def ensure_in_sys_path(
    path: Optional[Union[str, Path]],
    position: Optional[int] = None,
    resolve: bool = False,
    absolute: bool = True,
) -> None:
    """

    Ensures that a path is in sys.path, but avoids duplicates.
    Can also resolve and absolute paths for duplication.
    Does not clean the existing paths in sys.path

    :param path:
    :type path:
    :param position:
    :type position:
    :param resolve:
    :type resolve:
    :param absolute:
    :type absolute:
    :return:
    :rtype:
    """
    if path is None:  # may be the case if the supplied path is being solved programmatically
        warn("No path was supplied")
        return

    path = Path(path)

    if absolute:
        path = path.absolute()

    str_path = str(path)
    sys_path_snapshot = sys.path

    if resolve:
        sys_path_snapshot = [Path(p).resolve() for p in sys_path_snapshot]
        inclusion_test = path.resolve() in sys_path_snapshot
    else:
        inclusion_test = str_path in sys_path_snapshot

    if not inclusion_test:
        if position:
            sys.path.insert(position, str_path)
        else:
            sys.path.append(str_path)
    else:
        print(f"{path} is already in sys path")


def is_module_available(module: str) -> bool:
    """**Return True if module is available.**

    Parameters
    ----------
    module: str
          Name of the module to be checked.

    Returns
    -------
    bool
          True if installed.
    """
    return find_spec(module) is not None


def import_warning(module_name: str) -> None:
    """
    Inform the user that module has been imported,
    useful when repeated imports are heavy in the contexts of multiprocessing.
    Lets the user identify which file is reporting heavy loading and restructure code to avoid repeated importing

    :param module_name:
    :return:
    """
    from sys import modules

    if module_name in modules:
        warn(
            f"You already {module_name} had imported, consider restructuring your code to avoid repeated imports"
        )


def reimported_warning(module_name: str) -> None:
    """
    Just an idea

    :return:
    """
    raise NotImplemented
    # TODO: touch .lock file to system for module_name for a multiprocess warning if already exists,
    # delete it again once process is done?
    # context_wrapper maybe useful


if __name__ == "__main__":

    def _main() -> None:
        """
        :rtype: None
        """
        mod = "matplotlib"
        import_warning(mod)
        from matplotlib import pyplot

        import_warning(mod)
        pyplot.figure()

    def aisjdi():
        from copy import deepcopy

        s = deepcopy(sys.path)
        ensure_in_sys_path(Path(__file__).parent)
        s2 = sys.path
        print(s == s2, set(s2) - set(s), set(s) - set(s2), s2)

    def iajsd():
        from copy import deepcopy

        s = deepcopy(sys.path)
        clean_sys_path()
        s2 = sys.path
        print(s == s2, set(s2) - set(s), set(s) - set(s2), s2)

    def asuhdsaud():
        print(find_ancestral_relatives("queues"))

    # _main()
    # aisjdi()
    # iajsd()
    asuhdsaud()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/11/2019
           """

__all__ = ["timeit", "StopWatch"]

import contextlib
import functools
import time
import typing
from functools import wraps
from typing import Sequence, MutableMapping


def timeit(f: callable):
    """

    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(f)
    def wrapper(*args, **kwds):
        """

        :param args:
        :type args:
        :param kwds:
        :type kwds:
        :return:
        :rtype:"""
        start_time = time.time()
        result = f(*args, **kwds)
        elapsed_time = time.time() - start_time
        print(f"{f} took {elapsed_time:.3f} seconds to compute")
        return elapsed_time, result

    return wrapper


class StopWatch(contextlib.AbstractContextManager):
    r"""**Measure execution time of function.**

    Can be used as context manager or function decorator, perform checkpoints
    or display absolute time from measurements beginning.

    **Used as context manager**::

      with Timer() as timer:
          ... # your operations
          print(timer) # __str__ calls timer.time() internally
          timer.checkpoint() # register checkpoint
          ... # more operations
          print(timer.checkpoint()) # time since last timer.checkpoint() call

      ... # even more operations
      print(timer) # time taken for the block, will not be updated outside of it

    When execution leaves the block, timer will be blocked. Last checkpoint and time taken
    to execute whole block will be returned by `checkpoint()` and `time()` methods respectively.

    **Used as function decorator**::

      @Timer()
      def foo():
          return 42

      value, time = foo()

    Parameters
    ----------
    function : Callable, optional
          No argument function used to measure time. Default: time.perf_counter
    """

    def __init__(
        self,
        function: typing.Callable = time.perf_counter,
        auto_start_on_construction: bool = False,
        auto_start_on_enter: bool = True,
        auto_stop_on_exit: bool = True,
    ):
        self._stopped: bool = False
        self._started = False

        self._callable = function
        self._auto_start_on_construction = auto_start_on_construction
        self._auto_start_on_enter = auto_start_on_enter
        self._auto_stop_on_exit = auto_stop_on_exit

        self.start_time = 0
        self.new_time = 0
        self.previous_time = 0

        if self._auto_start_on_construction:
            self.start_timer()

        self.override_arithmetics()

    def override_arithmetics(self):
        """description"""

        def make_func(name):
            """description"""
            return lambda self, *args: getattr(self.since_start, name)(*args)

        arithmetics = (
            "add",
            "sub",
            "mul",
            "div",
            "truediv",
            "floordiv",
            "mod",
            "divmod",
            "pow",
        )

        methods = [
            "__invert__",
            "__neg__",
            "__pos__",
            "abs",
            "__round__",
            "__floor__",
            "__ceil__",
            "__int__",
            "__float__",
        ]
        methods.extend([f"__{n}__" for n in arithmetics])
        methods.extend([f"__r{n}__" for n in arithmetics])
        methods.extend([f"__i{n}__" for n in arithmetics])

        for name in methods:
            setattr(StopWatch, name, make_func(name))

    def start_timer(self):
        """description"""
        self._started = True
        self.start_time = self._callable()
        self.new_time = self.start_time
        self.previous_time = self.start_time

    def stop_timer(self):
        """description"""
        self.new_time = self._callable()
        self._stopped: bool = True

    @property
    def since_start(self):
        """**Time taken since the start of timer (measurements beginning).**

        Returns
        -------
        time-like
            Whatever `self.function() - self.function()` returns,
            usually fraction of seconds"""
        if not self._stopped and self._started:
            return self._callable() - self.start_time
        return self.new_time - self.start_time

    def tick(self):
        """**Time taken since last tick call.**

        If wasn't called before, it is the same as as Timer creation time (first call returns
        the same thing as `time()`)

        Returns
        -------
        time-like
            Whatever `self.function() - self.function()` returns,
            usually fraction of seconds"""
        if not self._stopped:
            if self._started:
                self.previous_time = self.new_time
                self.new_time = self._callable()
            else:
                self.start_timer()
        return self.new_time - self.previous_time

    def __call__(self, function):
        """
        decorator functionality

        :param function:
        :type function:
        :return:
        :rtype:"""

        @functools.wraps(function)
        def decorated(*args: Sequence, **kwargs: MutableMapping):
            """description"""
            self.start_timer()
            values = function(*args, **kwargs)
            self.stop_timer()
            return values, self.since_start

        return decorated

    def __enter__(self):
        if self._auto_start_on_enter:
            self.start_timer()
        return self

    def __exit__(self, *_, **__) -> bool:
        if self._auto_stop_on_exit:
            self.stop_timer()
        return False

    def __str__(self) -> str:
        return str(self.__repr__())

    def __repr__(self) -> int:
        return self.since_start


if __name__ == "__main__":
    a = StopWatch()
    print(f"Timer str rep: {a}")
    print(a.tick())
    print(a.tick())
    print(a // 2)
    print()

    with StopWatch(auto_start_on_enter=False) as timer1:
        print(timer1)  # __str__ calls timer.time() internally
        timer1.tick()  # register checkpoint
        print(timer1.tick())  # time since last timer.checkpoint() call

    print()
    with StopWatch() as timer4:
        print(timer4)  # __str__ calls timer.time() internally
        timer4.tick()  # register checkpoint
        print(timer4.tick())  # time since last timer.checkpoint() call

    print()
    print(timer4)  # time since start
    print(timer4.tick())  # time taken for the block, will not be updated outside of it
    print(timer4.tick())  # time taken for the block, will not be updated outside of it
    print(timer4)  # ime since start, will not be updated outside of it
    print()

    with StopWatch(auto_start_on_construction=True, auto_start_on_enter=False) as timer2:
        print(timer2)  # __str__ calls timer.time() internally
        print(timer2.tick())  # time since last timer.checkpoint() call
        print(timer2)

    @StopWatch()
    def foo() -> int:
        """
        :rtype: None
        """
        return 42

    value, time = foo()
    print(f"foo time: {time}, value: {value}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import inspect
from typing import Sequence, MutableMapping

from warg.decorators.kw_passing import (
    drop_unused_kws,
    passes_kws_to,
    super_init_pass_on_kws,
)

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
           """


def test_subclass_fully_qualified_no_args_or_kwargs():
    class BaseClass:
        """description"""

        def __init__(self, arg0, kwarg0=None, kwarg1=None):
            """

            :param arg0:
            :param kwarg0:
            :param kwarg1:"""
            self.arg0 = arg0
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_with_args_and_kwargs_on_subclasses():
    class BaseClass:
        """description"""

        def __init__(self, arg0, *args, kwarg0=None, kwarg1=None, **kwargs: MutableMapping):
            """

            :param arg0:
            :param args:
            :param kwarg0:
            :param kwarg1:
            :param kwargs:"""
            self.arg0 = arg0
            for key, val in enumerate(args):
                setattr(self, f"arg{key + 1}", val)
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1
            self.__dict__.update(kwargs)

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, *args, kwarg2=None, **kwargs: MutableMapping):
            super().__init__(arg0, *args, **kwargs)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws()
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, *args, kwarg2=None, **kwargs: MutableMapping):
            """

            :param arg0:
            :param arg1:
            :param arg2:
            :param args:
            :param kwarg2:
            :param kwargs:"""
            super().__init__(arg0, *args, **kwargs)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_subclass_with_kwargs():
    class BaseClass:
        """description"""

        def __init__(self, arg0, kwarg0=None, kwarg1=None):
            """

            :param arg0:
            :param kwarg0:
            :param kwarg1:"""
            self.arg0 = arg0
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None, **kwargs: MutableMapping):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None, **kwargs: MutableMapping):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_subclass_with_args():
    class BaseClass:
        """description"""

        def __init__(self, arg0, kwarg0=None, kwarg1=None):
            """

            :param arg0:
            :param kwarg0:
            :param kwarg1:"""
            self.arg0 = arg0
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, *args, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, *args, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_base_with_kwargs():
    class BaseClass:
        """description"""

        def __init__(self, arg0, kwarg0=None, kwarg1=None, **kwargs: MutableMapping):
            """

            :param arg0:
            :param kwarg0:
            :param kwarg1:"""
            self.arg0 = arg0
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_base_with_args():
    class BaseClass:
        """description"""

        def __init__(self, arg0, *args, kwarg0=None, kwarg1=None):
            """

            :param arg0:
            :param kwarg0:
            :param kwarg1:"""
            self.arg0 = arg0
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, kwarg0=kwarg0)
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_base_with_args_and_mock_empty_dict():
    class BaseClass:
        """description"""

        def __init__(self, arg0, *args, kwarg0=None, kwarg1=None):
            """

            :param arg0:
            :param kwarg0:
            :param kwarg1:"""
            self.arg0 = arg0
            self.kwarg0 = kwarg0
            self.kwarg1 = kwarg1

    class SubClass0(BaseClass):
        """description"""

        @passes_kws_to(BaseClass.__init__)
        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, (), kwarg0=kwarg0, **{})
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    @super_init_pass_on_kws
    class SubClass1(BaseClass):
        """description"""

        def __init__(self, arg0, arg1, arg2, kwarg0=0, kwarg2=None):
            super().__init__(arg0, (), kwarg0=kwarg0, **{})
            self.arg1 = arg1
            self.arg2 = arg2
            self.kwarg2 = kwarg2

    print(inspect.signature(SubClass0.__init__))
    print(inspect.signature(SubClass1.__init__))

    print(vars(SubClass0(1, 1, 1, kwarg0=52)))
    print(vars(SubClass1(2, 2, 1, kwarg0=52)))


def test_chaining_arbitrary_kwargs():
    def b(c, f, *args, d=None, **kwargs: MutableMapping):
        """description"""
        pass

    def l(im_here=None, **kwargs: MutableMapping):
        """description"""
        pass

    @passes_kws_to(b)
    @passes_kws_to(l)
    def a(e, *args: Sequence, **kwargs: MutableMapping):
        """description"""
        b(1, 2, **kwargs)
        l(**kwargs)

    print(inspect.signature(a))
    a(1, d=None, im_here=2)


def test_chaining_arbitrary_kwargs_keep():
    def b(c, f, *args, d=None, **kwargs: MutableMapping):
        """description"""
        pass

    def l(im_here=None, **kwargs: MutableMapping):
        """description"""
        pass

    @passes_kws_to(b, keep_from_var_kw=True)
    @passes_kws_to(l, keep_from_var_kw=True)
    def a(e, *args: Sequence, **kwargs: MutableMapping):
        """description"""
        b(1, 2, **kwargs)
        l(**kwargs)

    print(inspect.signature(a))
    a(1, d=None, im_here=2)


def test_chaining_no_keep():
    def b(c, f, *args, d: float = None, **kwargs: MutableMapping):
        """description"""
        pass

    def l(a, im_here: str = None, **kwargs: MutableMapping):
        """

        :param a:
        :param im_here: Nice to meet you
        :param kwargs:
        :return:"""
        pass

    @passes_kws_to(b)
    @passes_kws_to(l)
    def a1(e: int, *args: Sequence, **kwargs: MutableMapping):
        """

        :param e:
        :param args:
        :param kwargs:
        :return:"""
        b(1, 2, **kwargs)
        l(1, **kwargs)

    print("a1", inspect.signature(a1))

    @passes_kws_to(b, l)
    def a(e: int, *args: Sequence, **kwargs: MutableMapping):
        """description"""
        pass
        b(1, 2, **kwargs)
        l(1, **kwargs)

    print(inspect.signature(a))
    a(1, d=None, im_here=None)


def test_chaining_no_keep_composed_drop_kws():
    def b(c, f, *args, d: float = None, **kwargs: MutableMapping):
        """description"""
        pass

    def l(a, im_here: str = None, **kwargs: MutableMapping):
        """

        :param a:
        :param im_here: Nice to meet you
        :param kwargs:
        :return:"""
        pass

    @passes_kws_to(b)
    @passes_kws_to(l)
    def a1(e: int, *args: Sequence, **kwargs: MutableMapping):
        """

        :param e:
        :param args:
        :param kwargs:
        :return:"""
        b(1, 2, **kwargs)
        l(1, **kwargs)

    print("a1", inspect.signature(a1))

    @drop_unused_kws
    @passes_kws_to(b, l)
    def a(e: int, *args: Sequence, **kwargs: MutableMapping):
        """description"""
        pass
        b(1, 2, **kwargs)
        l(1, **kwargs)

    print(inspect.signature(a))
    a(1, d=None, im_here=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/5/22
           """

__all__ = ["export"]

import sys


# @export # Sadly can not be used here as it not declared yet :/ ;)
def export(fn):
    """

    A decorator that exports function in the containing modules all in not already existent.

    works but code analysis tools usually complains

        :param fn:
        :type fn:
        :return:
        :rtype:
    """
    mod = sys.modules[fn.__module__]
    if hasattr(mod, "__all__"):
        mod.__all__.append(fn.__name__)
    else:
        mod.__all__ = [fn.__name__]
    return fn



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 07-05-2021
           """

import os
import subprocess
from pathlib import Path

__all__ = ["latest_file", "exist_any_extension", "system_open_path"]

from warg.os_utilities.os_platform import is_windows, is_mac, has_x_server


def system_open_path(path: Path, *, verbose: bool = False) -> None:
    """
    Use system defaults for opening path/uris

    :param path:
    :type path:
    :param verbose:
    :type verbose:
    :return:
    :rtype:
    """
    if has_x_server():
        if verbose:
            print(f"Opening the directory ({path}) using the systems default file manager")

        directory = str(path)

        if is_windows():
            subprocess.Popen(["start", directory], shell=True)

        elif is_mac():
            subprocess.Popen(["open", directory])

        else:
            # try:
            subprocess.Popen(["xdg-open", directory])
            # except OSError:
    else:
        print("Target display not set")


def latest_file(
    directory: Path,
    extension: str = "",
    *,
    recurse: bool = False,
    raise_on_failure: bool = True,
) -> Path:
    """

    :param directory:
    :param extension:
    :param recurse:
    :param raise_on_failure:
    :return:
    """
    a = f"*{extension}"
    if recurse:
        path_gen = directory.rglob(a)
    else:
        path_gen = directory.glob(a)
    list_of_files = list(path_gen)
    if len(list_of_files) == 0:
        msg = f"Found no previous files with extension {extension} in {directory}"
        if raise_on_failure:
            raise FileNotFoundError(msg)
        print(f"{msg}, returning None!")
        return
    return max(list_of_files, key=os.path.getctime)  # USES CREATION TIME


def exist_any_extension(p: Path) -> bool:
    """
    If any file with that stem exist in parent directory, return True.

    :param p:
    :type p:
    :return:
    :rtype:
    """
    for _ in p.parent.glob(f"{p.stem}.*"):
        return True
    return False


if __name__ == "__main__":
    print(latest_file(Path(__file__).parent, recurse=True))
    print(exist_any_extension(Path(__file__)))
    print(exist_any_extension(Path.cwd() / "__init__.py"))
    print(exist_any_extension(Path.cwd() / "__init__"))
    print(exist_any_extension(Path.cwd() / "__init__.test"))
    print(exist_any_extension(Path.cwd() / "__init___.py"))



from typing import Mapping, Iterable, Hashable, Dict

__all__ = [
    "invert_mapping",
    "invert_dict",
    "AppendingDict",
    "pivot_dict_object",
    "pivot_dict",
]


def append_to_dict(d: Dict, key, value) -> Dict:
    """

    :param d:
    :type d:
    :param key:
    :type key:
    :param value:
    :type value:
    :return:
    :rtype:
    """
    d.setdefault(key, []).append(value)
    return d


class AppendingDict(Dict):  # appending_dict = collections.defaultdict(list)
    def __setitem__(self, key, value):
        # self.setdefault(key, []).append(value)
        # append_to_dict(self, key, value)
        if key in self:
            self[key].append(value)
        else:
            super().__setitem__(key, [value])


def recurse_mapping(a: Mapping, call: callable = print) -> None:
    for k, v in a.items():
        if isinstance(v, Mapping):
            recurse_mapping(v)
        call(v)


def invert_mapping(m: Mapping) -> Mapping:
    """
    Invert a mapping

    if a mapping does not have duplicate hashable values, then this is the same as invert_dict, otherwise values in
    new_m are tuples of keys with duplicate values
    :return: :rtype:
    """

    new_m = type(m)()

    for k, v in m.items():
        if not isinstance(v, Hashable):
            raise TypeError(f"values must be hashable, was {type(v), v}, for key {k}")
        if v in new_m:
            if isinstance(new_m[v], Iterable):
                new_m[v] = (*new_m[v], k)
            else:
                new_m[v] = [new_m[v], k]
        else:
            new_m[v] = k
    return new_m


def invert_dict(d: Mapping) -> Dict:
    """
    Invert a dict

    :param d:
    :type d:
    :return:
    :rtype:
    """
    return dict(((v, k) for k, v in d.items()))


def pivot_dict(d: Dict, key) -> Dict:
    """
    pivot_key -> pivot_value

    :param d:
    :param key:
    :return:
    :rtype:
    """
    return dict(((v[key], k) for k, v in d.items()))


def pivot_dict_object(d: Dict, key) -> Dict:
    """
    pivot_key -> pivot_value for object attributes

    :param d:
    :param key:
    :return:
    :rtype:
    """
    return dict(((getattr(v, key), k) for k, v in d.items()))


if __name__ == "__main__":
    print(invert_mapping({"a": 1, "b": 2}))

    print(invert_mapping({"a": 2, "b": 2, "c": 3, "d": 4}))
    print(invert_dict({"a": 2, "b": 2, "c": 3, "d": 4}))

    # print(pivot_dict_object({"a": 2, "b": 2, "c": 3, "d": 4}, "id"))



from typing import Any

__all__ = ["dist_is_editable", "package_is_editable"]

import pkg_resources


def dist_is_editable(dist: Any) -> bool:
    """
    Return True if given Distribution is an editable installation."""
    import sys
    from pathlib import Path

    for path_item in sys.path:
        egg_link = Path(path_item) / f"{dist.project_name}.egg-link"
        if egg_link.is_file():
            return True
    return False


def package_is_editable(package_name: Any) -> bool:
    distributions = {v.key: v for v in pkg_resources.working_set}
    if package_name in distributions:
        distribution = distributions[package_name]
        return dist_is_editable(distribution)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 8/30/22
           """

__all__ = ["is_excluded", "is_python_package", "is_python_module", "negate"]

import re
from functools import wraps
from pathlib import Path
from typing import Callable, Sequence, MutableMapping, Union, Any


def is_python_module(path: Path) -> bool:
    """
    Check if path is a python module
    """
    path = Path(path)

    return path.is_file() and path.suffix == ".py"


def is_python_package(path: Path) -> bool:
    """
    Check if path is a python package
    """
    path = Path(path)
    return path.is_dir() and (path / "__init__.py").exists()


def is_excluded(path: Union[Path, str], exclude_pattern: str = "[Ee]xcluded*"):
    """
    is exclude by common pattern or gitignore

    #TODO: Look up nearest parent git ignore and interpret

    :param exclude_pattern:
    :type exclude_pattern:
    :param path:
    :type path:
    :return:
    :rtype:
    """

    if len(re.findall(exclude_pattern, str(path))) > 0:
        return True
    return False


def negate(f: Callable) -> Callable:
    """
    Negate a function return
    """

    assert isinstance(f, Callable), "ensure you did not call the callable with parameters directly"

    @wraps(f)
    def wrapper(*args: Sequence, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return not f(*args, **kwargs)

    return wrapper


if __name__ == "__main__":
    # print(negate(is_excluded("/iods/excludes/osad.py")))
    print(negate(is_excluded)("/iods/excludes/osad.py"))
    print(is_excluded("/s/excludes/a.py"))
    print(is_excluded("/s/exclud/a.py"))
    print(is_excluded("/s/exclude/a.py"))
    print(is_excluded("/s/Exclude/a.py"))
    print(is_excluded("/s/excluded/a.py"))
    print(is_excluded("/s/Excluded/a.py"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09-09-2020
           """

__all__ = [
    "map_value_product",
    "map_product",
    "map_permutations",
    "map_combinations",
    "map_combinations_with_replacement",
]

import itertools
from typing import Any, Mapping, Tuple


def map_value_product(dicts: Mapping) -> Any:
    """description"""
    return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))


def map_product(dicts: Mapping, repeat: int = 2) -> Any:
    """description"""
    yield from zip(
        itertools.product(dicts.keys(), repeat=repeat),
        itertools.product(dicts.values(), repeat=repeat),
    )


def map_permutations(dicts: Mapping, repeat: int = 2) -> Tuple:
    """description"""
    yield from zip(
        itertools.permutations(dicts.keys(), repeat),
        itertools.permutations(dicts.values(), repeat),
    )


def map_combinations(dicts: Mapping, repeat: int = 2) -> Tuple:
    """description"""
    yield from zip(
        itertools.combinations(dicts.keys(), repeat),
        itertools.combinations(dicts.values(), repeat),
    )


def map_combinations_with_replacement(dicts: Mapping, repeat: int = 2) -> Tuple:
    """description"""
    yield from zip(
        itertools.combinations_with_replacement(dicts.keys(), repeat),
        itertools.combinations_with_replacement(dicts.values(), repeat),
    )


if __name__ == "__main__":

    def asdijha() -> None:
        """
        :rtype: None
        """
        from warg import NOD

        a = NOD(a=[1], b=[4], c=[8])
        print(f"ValueMapProduct{str(list(map_value_product(a.as_dict())))}")
        print(f"MapProduct{str(list(map_product(a.as_dict())))}")
        print(f"map_combinations{str(list(map_combinations(a.as_dict())))}")
        print(f"map_permutations{str(list(map_permutations(a.as_dict())))}")
        print(f"map_combinations_with_replacement{str(list(map_combinations_with_replacement(a.as_dict())))}")

    def asdijhsadasdad() -> None:
        """
        :rtype: None
        """
        from warg import NOD

        a = NOD(a=[1, 2, 8], b=[4, 3, 99])
        print(f"ValueMapProduct{str(list(map_value_product(a.as_dict())))}")
        print(f"MapProduct{str(list(map_product(a.as_dict())))}")
        print(f"map_combinations{str(list(map_combinations(a.as_dict())))}")
        print(f"map_permutations{str(list(map_permutations(a.as_dict())))}")
        print(f"map_combinations_with_replacement{str(list(map_combinations_with_replacement(a.as_dict())))}")

    asdijha()
    # asdijhsadasdad()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Sequence, MutableMapping

from warg import drop_unused_kws
from warg.metas.post_init import PostInit

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
           """


def test_post_init_class():
    class MyTestingClass(metaclass=PostInit):
        """
        class with the metaclass passed as an argument"""

        @drop_unused_kws
        def __init__(self, *args: Sequence, **kwargs: MutableMapping):
            print(kwargs)

        def __post_init__(self, *args: Sequence, **kwargs: MutableMapping):
            print(args, kwargs)

        def __call__(self, *args: Sequence, **kwargs: MutableMapping):
            print("a")

    a = MyTestingClass("asdc", kas=2)

    a()


def test_post_init_no_kws_class():
    class MyTestingClass(metaclass=PostInit):
        """
        class with the metaclass passed as an argument"""

        @drop_unused_kws
        def __init__(self, *args: Sequence):
            print("Init class")

        @drop_unused_kws
        def __post_init__(self, *args: Sequence):
            print(args)

        def __call__(self, *args: Sequence, **kwargs: MutableMapping):
            print("a")

    a = MyTestingClass("asdc", kas=2)

    a()


def test_no_post_init_class():
    class MyTestingClass(metaclass=PostInit):
        """
        class with the metaclass passed as an argument"""

        def __init__(self):
            print("Init class")

        def __call__(self, *args: Sequence, **kwargs: MutableMapping):
            print("a")

    a = MyTestingClass()

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/03/2020
           """

from itertools import cycle
from typing import Iterable


def busy_indicator(
    *,
    stream: callable = print,
    indicator_interval: int = 1,
    phases: Iterable[str] = ("◑", "◒", "◐", "◓"),
) -> Iterable:
    """
    You can choose arbitrary phases like ['|','/','-','\\']

    :param stream:
    :type stream:
    :param indicator_interval:
    :type indicator_interval:
    :param phases:
    :type phases:
    :return:
    :rtype:"""

    phases = cycle(phases)
    i = 0
    while True:
        if i % indicator_interval == 0:
            stream(f"{next(phases)}", end="\r", flush=True)
        yield i
        i += 1


if __name__ == "__main__":

    def iasd():
        """description"""
        import time

        for i in busy_indicator(indicator_interval=10):
            time.sleep(0.1)
            if i > 100:
                break

    iasd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/11/2019
           """

__all__ = ["pre_decorate", "post_decorate"]


def pre_decorate(method: callable, *callables: callable):
    def pre_call_func(self: object = None, *args, **kwargs):
        for c in callables:
            c(self, *args, **kwargs)
        return method(self, *args, **kwargs)

    return pre_call_func


def post_decorate(method: callable, *callables: callable):
    def post_call_func(self: object = None, *args, **kwargs):
        res = method(self, *args, **kwargs)
        for c in callables:
            c(self, *args, res=res, **kwargs)
        return res

    return post_call_func


if __name__ == "__main__":

    def juahsdu():
        def c(d):
            print(d)
            return f"c_{d}"

        a = pre_decorate(c, lambda *args, **kwargs: print(f"pre {args, kwargs}"))
        b = post_decorate(c, lambda *args, **kwargs: print(f"post {args, kwargs}"))

        print(a("yo"))
        print(b("bro"))

    juahsdu()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02/01/2020
           """

__all__ = [
    "is_positive_and_mod_zero",
    "is_zero_or_mod_zero",
    "is_none_or_zero_or_negative",
    "is_zero_or_mod_below",
    "is_none_or_zero_or_negative_or_mod_zero",
    "xor",
    "xnor",
    "nand",
]

import math
from typing import Any, Optional

from warg import Number
from warg.decorators import drop_unused_kws, passes_kws_to


@drop_unused_kws
def is_positive_and_mod_zero(
    mod: Optional[Number],
    counter: int,
    *,
    ret: Any = True,
    alt: Any = False,
    residual_printer: callable = None,
) -> Any:
    """

    test if mod is positive
    then test if counter % mod is 0
    if both tests are true return ret
    else return alt


    :param residual_printer:
    :type residual_printer:
    :param mod:
    :param counter:
    :param ret:
    :param alt:
    :return:"""

    if mod == 0:
        if residual_printer is not None:
            residual_printer(math.inf)
        return alt

    m = counter % mod
    if residual_printer is not None:
        residual_printer(m)

    return ret if (mod > 0 and (m == 0)) else alt


@drop_unused_kws
def is_zero_or_mod_below(
    mod: Optional[Number],
    below: Number,
    counter: int,
    *,
    ret: Any = True,
    alt: Any = False,
    residual_printer: callable = None,
) -> Any:
    """

    test if mod is zero or if counter % mod is 0
    if any of the tests are true return ret
    else return alt


    :param residual_printer:
    :type residual_printer:
    :param below:
    :type below:
    :param mod:
    :param counter:
    :param ret:
    :param alt:
    :return:"""
    if mod == 0:
        if residual_printer is not None:
            residual_printer(0)
        return ret

    m = counter % mod
    if residual_printer is not None:
        residual_printer(m - below)

    return ret if (m < below) else alt


@drop_unused_kws
def is_zero_or_mod_zero(
    mod: Optional[Number],
    counter: int,
    *,
    ret: Any = True,
    alt: Any = False,
    residual_printer: callable = None,
) -> Any:
    """

    test if mod is zero or if counter % mod is 0
    if any of the tests are true return ret
    else return alt


    :param residual_printer:
    :type residual_printer:
    :param mod:
    :param counter:
    :param ret:
    :param alt:
    :return:"""

    if mod == 0:
        if residual_printer is not None:
            residual_printer(0)
        return ret

    m = counter % mod
    if residual_printer is not None:
        residual_printer(m)

    return ret if (m == 0) else alt


@drop_unused_kws
def is_none_or_zero_or_negative(obj: Optional[Number], residual_printer: callable = None) -> bool:
    """

    :param residual_printer:
    :type residual_printer:
    :param obj:
    :return:"""
    is_none = obj is None
    is_negative = False
    if isinstance(obj, (int, float)):
        is_negative = obj <= 0

    if is_none:
        if residual_printer is not None:
            residual_printer(0)
    else:
        if residual_printer is not None:
            residual_printer(obj)

    return is_none or is_negative


@passes_kws_to(is_zero_or_mod_zero)
def is_none_or_zero_or_negative_or_mod_zero(mod: Optional[Number], counter: int, **kwargs) -> bool:
    """

    :param mod:
    :param counter:
    :param kwargs:
    :return:"""
    return is_none_or_zero_or_negative(mod, **kwargs) or is_zero_or_mod_zero(mod, counter, **kwargs)


def xor(a: bool, b: bool) -> bool:
    """

    :param a:
    :param b:
    :return:"""
    return a ^ b


def xnor(a: bool, b: bool) -> bool:
    """

    :param a:
    :param b:
    :return:"""
    return not xor(a, b)


def nand(a: bool, b: bool) -> bool:
    """

    :param a:
    :param b:
    :return:"""
    return not (a and b)


if __name__ == "__main__":
    assert is_zero_or_mod_below(5, 3, 7) == True
    assert is_zero_or_mod_below(5, 2, 4) == False
    for i in range(9):
        print(is_zero_or_mod_zero(1, i))
    for i in range(9):
        print(is_zero_or_mod_zero(2, i))

    print()
    for i in range(8):
        print(is_none_or_zero_or_negative_or_mod_zero(None, i, residual_printer=print))

    for i in range(8):
        print(is_none_or_zero_or_negative_or_mod_zero(True, 4 - i, residual_printer=print))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 9/14/19
           """

from _pytest.capture import CaptureFixture


def test_print(capsys: CaptureFixture[str]) -> None:
    """Correct my_name argument prints"""
    text = "hello"
    err = "world"
    print(text)
    sys.stderr.write("world")
    captured = capsys.readouterr()
    assert text in captured.out
    assert err in captured.err


if __name__ == "__main__":
    test_print()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          This file is not for general use. Ode to python

           Created on 13/06/2020
           """

__all__ = []


def this_is_barely_legal():
    try:
        try:
            try:
                try:
                    try:
                        try:
                            try:
                                try:
                                    try:
                                        try:
                                            try:
                                                try:
                                                    try:
                                                        try:
                                                            try:
                                                                try:
                                                                    try:
                                                                        try:
                                                                            try:
                                                                                """
                                                                                try:
                                                                                  try:
                                                                                    try:
                                                                                      try:
                                                                                        try:
                                                                                          try:
                                                                                            try:
                                                                                              try:
                                                                                                try:
                                                                                                    ...
                                                                                                except:
                                                                                                    ...
                                                                                              except:
                                                                                                  ...
                                                                                            except:
                                                                                                ...
                                                                                          except:
                                                                                              ...
                                                                                        except:
                                                                                            ...
                                                                                      except:
                                                                                          ...
                                                                                    except:
                                                                                        ...
                                                                                  except:
                                                                                      ...
                                                                                except:
                                                                                  ...
                                                                                """
                                                                            except:
                                                                                ...
                                                                        except:
                                                                            ...
                                                                    except:
                                                                        ...
                                                                except:
                                                                    ...
                                                            except:
                                                                ...
                                                        except:
                                                            ...
                                                    except:
                                                        ...
                                                except:
                                                    ...
                                            except:
                                                ...
                                        except:
                                            ...
                                    except:
                                        ...
                                except:
                                    ...
                            except:
                                ...
                        except:
                            ...
                    except:
                        ...
                except:
                    ...
            except:
                ...
        except:
            ...
    except:
        ...


if __name__ == "__main__":
    # from __future__ import braces # TODO: Try disable comment

    # noinspection PyUnresolvedReferences
    import __hello__

    print()
    import this

    print()
    # noinspection PyUnresolvedReferences
    import antigravity

    print()
    print(this.i, this.d, this.c, this.s)
    print()

    # noinspection PyUnresolvedReferences
    # import that  # pip install that

    this_is_barely_legal()



import itertools
from math import sin, cos
from typing import Iterable

from warg import Number

__all__ = ["sin_gen", "cos_gen", "loop"]

loop = itertools.cycle


def sin_gen(iterable: Iterable[Number], magnitude: Number = 1) -> Iterable[Number]:
    """

    :param magnitude:
    :type magnitude:
    :param iterable:
    :type iterable:"""
    for a in iterable:
        yield sin(a) * magnitude


def cos_gen(iterable: Iterable, magnitude: Number = 1) -> Iterable[Number]:
    """

    :param magnitude:
    :type magnitude:
    :param iterable:
    :type iterable:"""
    for a in iterable:
        yield cos(a) * magnitude


if __name__ == "__main__":

    def assda():
        """description"""
        import numpy

        for i in cos_gen(numpy.arange(0, 100, 0.1)):
            print(i)

    assda()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/08/2020
           """

__all__ = []

from pathlib import Path


def test_import():
    import warg

    print(warg.__version__)


def test_find_nearest_ancestral_relative():
    from warg import find_nearest_ancestral_relative

    assert (
        find_nearest_ancestral_relative(
            "scripts",
            top_level=Path(__file__).parent.parent,
            context=Path(__file__),
            ancestral_levels=2,
            descendant_levels=1,
        )
        == Path(__file__).parent.parent
    )


def test_find_ancestral_relatives():
    from warg import find_ancestral_relatives

    assert find_ancestral_relatives(
        "warg",
        top_level=Path(__file__).parent.parent,
        context=Path(__file__),
        ancestral_levels=1,
        descendant_levels=1,
    ) == [Path(__file__).parent.parent]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

__doc__ = """description"""
__all__ = ["to_british_english", "deamericanise"]

from types import MappingProxyType
from typing import Mapping

default_rules = MappingProxyType(
    {
        "ize": "ise",
        "yze": "yse",
        "iza": "isa",
        "aluminum": "aluminium"
        # 'se': 'ce',
        # 'og': 'ogue',
    }
)


def to_british_english(text: str, rules: Mapping = default_rules) -> str:
    """

    :param text:
    :type text:
    :param rules:
    :type rules:
    :return:
    :rtype:
    """
    for r in rules.items():
        text = text.replace(*r)
    return text


def deamericanise(text: str) -> str:
    """
    Naively exchanges 'z' in english texts

    convert to 'British English'

    :return:
    :rtype:
    """
    return to_british_english(text)


if __name__ == "__main__":
    print(
        deamericanise(
            "I analyzed websites in order to recognize the correct spelling of international organizations"
        )
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18-01-2021
           """

from enum import Enum
from typing import Iterable, Any

__all__ = ["FilterModeEnum", "symbol_filter"]


# from sorcery import assigned_names # PURE!


class FilterModeEnum(Enum):
    """
    Filter Mode
    """

    # exclude_postfix, exclude_prefix, exclude_fully = assigned_names()
    # TODO: Include variants
    exclude_postfix, exclude_prefix, exclude_fully = (
        "exclude_postfix",
        "exclude_prefix",
        "exclude_fully",
    )


def symbol_filter(
    string_stream: Iterable,
    symbol: str = "#",
    *,
    exclusion_mode: FilterModeEnum = FilterModeEnum.exclude_postfix,
) -> Any:
    """description"""
    if exclusion_mode == FilterModeEnum.exclude_fully:
        yield from filter(lambda s: symbol not in s, string_stream)
    elif exclusion_mode == FilterModeEnum.exclude_postfix or exclusion_mode == FilterModeEnum.exclude_prefix:
        selector = 0
        if exclusion_mode == FilterModeEnum.exclude_prefix:
            selector = -1
        for s in string_stream:
            raw = s.split(symbol)[selector].strip()
            if raw:
                yield raw
    else:
        raise NotImplemented(f"{exclusion_mode} mode not supported")


if __name__ == "__main__":

    def asijsda() -> None:
        """
        :rtype: None
        """
        strings = [
            " # aasd # sad ",
            " faojasasd # oiwaos ",
            " okjasifj  oajsidw2 ",
            " 12 329#9213",
        ]
        for s in symbol_filter(strings):
            print(s)

        print(" ")

        for i, s in enumerate(symbol_filter(strings, exclusion_mode=FilterModeEnum.exclude_fully)):
            print(i, s)

        print(" ")

        for s in symbol_filter(strings, exclusion_mode=FilterModeEnum.exclude_prefix):
            print(s)

    asijsda()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/12/2019
           """

from warg.data_structures.auto_dict import (
    AutoDict,
    sanitise_auto_dict,
    recursive_default_dict_print,
)


def test_a():
    ad = AutoDict()

    ad["b"]["b"]["c"]["b"]["c"] = {}
    ad["c"] = 1
    ad["cd"]["v"] = 412
    ad["qc"]["d"] = []
    ad["cf"]["b6"] = None
    ad["cf"]["1"] = None

    a = ad["b"]["b"]["b"]["b"]
    b = ad["b"]["a"]["gf"]["c"]
    c = ad["cd"]["v"]

    print(a, b, c)

    print(ad)

    recursive_default_dict_print(ad)

    d = sanitise_auto_dict(ad)
    try:
        a = d["b"]["b"]["b"]["b"]
    except KeyError as k:
        assert isinstance(k, KeyError)

    try:
        b = d["b"]["a"]["gf"]["c"]
    except KeyError as k:
        assert isinstance(k, KeyError)

    c = d["cd"]["v"]

    assert c == 412

    print(d)


if __name__ == "__main__":
    test_a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28-01-2021
           """

from warg import get_first_arg_name, identity


def test_ausdh3():
    from typing import Any

    def some_func(a: Any) -> None:
        """description"""
        print(get_first_arg_name("some_func", verbose=True))

    some_func(print(2, sep="-"))


def test_ausd2h3():
    from typing import Any

    def some_func(a: Any) -> None:
        """description"""
        print(get_first_arg_name("some_func", verbose=True))

    some_func(identity(2))


def test_ausd2h34():
    from typing import Any

    def some_func(a: Any) -> None:
        """description"""
        print(get_first_arg_name("some_func", verbose=True))

    asd = 2
    some_func(identity(asd))


def test_ausd2h3213():
    from typing import Any

    class Ac:
        class Bc:
            @staticmethod
            def c(d):
                """description"""
                pass

    def some_func(a: Any) -> None:
        """description"""
        print(get_first_arg_name("some_func", verbose=True))

    some_func(Ac.Bc.c(2))


if __name__ == "__main__":
    test_ausdh3()
    test_ausd2h3()
    test_ausd2h3213()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/03/2020
           """
__all__ = ["PropertySettings"]

from collections.abc import Mapping
from typing import Dict, MutableMapping, Any

from warg import NOD


class PropertySettings(
    # Mapping
):
    """description"""

    # def __getitem__(self, k):
    #  return  self.__getattribute__(k)

    # def __len__(self) -> int:
    #  return len(self.__crystallise__())

    # raise_exception_on_none = False
    raise_exception_non_exist_property = True

    def __init__(self, **kwargs: MutableMapping):
        for k, v in kwargs.items():
            self.__setattr__(k, v)

        for setting in dir(self):
            if not setting.startswith("_"):
                try:
                    self.__getattr__(setting)
                except KeyError:
                    self.__setattr__(setting, None)

    def __clear__(self):
        for setting in dir(self):
            if not setting.startswith("_"):
                self.__setattr__(setting, None)

    def __setattr__(self, key: str, value: Any):
        assert not key.startswith("_"), f"{key} is not allowed"
        if PropertySettings.raise_exception_non_exist_property and not hasattr(self, key):
            raise ValueError(f"Property setting {key} does not exist , available settings {self}")
        # self.__getattr__(key)
        super().__setattr__(key, value)

    def __getattr__(self, item):
        assert not item.startswith("_"), f"{item} is not allowed"
        try:
            val = super().__getattribute__(item)
            # if PropertySettings.raise_exception_on_none and not val:
            #  raise ValueError(f'Property {item} is {val}')
            return val
        except AttributeError as a:
            a = type(a)(str(a) + f", available settings {self}")
            raise a

    def __iter_keys__(self):
        for setting in dir(self):
            if not setting.startswith("_"):
                yield setting

    def __str__(self) -> str:
        return self.__repr__()

    def __contains__(self, item: Any):
        return hasattr(self, item)

    # def __dir__(self) -> Iterable[str]:
    #  return self.__crystallise__().keys()

    def __repr__(self) -> str:
        settings_dict = {}
        for setting in dir(self):
            if not setting.startswith("_"):
                try:
                    settings_dict[setting] = self.__getattr__(setting)
                except KeyError:
                    settings_dict[setting] = None

        return str(settings_dict)

    def __iter__(self):
        available_settings = []
        for setting in dir(self):
            if not setting.startswith("_"):
                available_settings.append(setting)

        return iter(available_settings)

    def __getitem__(self, item: Any):
        return self.__getattr__(item)

    def __to_dict__(self) -> Dict:
        return self.__crystallise__().as_dict()

    def __crystallise__(self) -> NOD:
        return NOD({k: getattr(self, k) for k in self})

    def __from_mapping__(self, mapping: Mapping) -> None:
        for k, v in mapping.items():
            setattr(self, k, v)

    def __from_dict__(self, d: Dict) -> None:
        self.__from_mapping__(d)


if __name__ == "__main__":
    a = PropertySettings()

    print({**a.__crystallise__()})
    assert not "h" in a



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 9/9/22
           """

__all__ = ["PrivateAttributeMixin"]


class PrivateAttributeMixin:
    def __getitem__(self, key):
        if key.startswith("_"):
            return None
        return getattr(self, key, None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09-02-2021
           """
__all__ = ["IgnoreInterruptSignal", "LambdaContext"]

import contextlib
import signal

from warg import AlsoDecorator


class LambdaContext(contextlib.AbstractContextManager):
    def __init__(self, callable_):
        self.callable_ = callable_

    def __enter__(self):
        return self.callable_

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    def __getattr__(self, item):
        return getattr(iter, self.callable_)

    def __getitem__(self, item):
        return self.callable_[item]

    def __call__(self, *args, **kwargs):
        return self.callable_


class IgnoreInterruptSignal(contextlib.AbstractContextManager, AlsoDecorator):
    """description"""

    def __enter__(self) -> bool:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # signal.getsignal() No sideeffect options for exit
        return True

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        signal.signal(signal.SIGINT, signal.SIG_DFL)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/07/2020
           """

__all__ = ["ContextWrapper", "NopContext"]

import contextlib
import inspect
from typing import Callable, Sequence


class NopContext(contextlib.AbstractContextManager):
    def __enter__(self):
        return

    def __exit__(self, exc_type, exc_val, exc_tb):
        return


class ContextWrapper(contextlib.AbstractContextManager):
    """
    Allows for conditional application of contexts, if uninstantiated context manager classes are passed no arguments is supplied in construction.
    if disabled None is returned
    if enabled return of context manager is propagated"""

    def __init__(
        self,
        context_manager: callable,
        enabled: bool,
        construction_args: Sequence = (),
        construction_kwargs=None,
    ):
        if construction_kwargs is None:
            construction_kwargs = {}

        self._context_manager = context_manager
        self._enabled = enabled
        self._construction_args = construction_args
        self._construction_kwargs = construction_kwargs

    def __enter__(self):
        if self._enabled:
            if inspect.isclass(self._context_manager) or isinstance(self._context_manager, Callable):
                self._context_manager = self._context_manager(
                    *self._construction_args, **self._construction_kwargs
                )
            if hasattr(self._context_manager, "__enter__"):
                return self._context_manager.__enter__()
            else:
                raise NotImplementedError(f"{self._context_manager} does not implement __enter__")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._enabled:
            return self._context_manager.__exit__(exc_type, exc_val, exc_tb)


if __name__ == "__main__":

    class SampleContextManager:
        """description"""

        def __init__(self, message="Hello World"):
            self._message = message

        def __enter__(self):
            print(self._message)

        def __exit__(self, exc_type, exc_val, exc_tb):
            print(not self._message)  # False ;)

    def main() -> None:
        """
        :rtype: None
        """
        with ContextWrapper(SampleContextManager(), True):
            print("with enabled")

        print()
        with ContextWrapper(SampleContextManager(), False):
            print("with disabled")

        print()
        with ContextWrapper(SampleContextManager, True):
            print("with enabled, uninstantiated")

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-01-2021
           """

__all__ = ["ArgIdentifier", "get_arg_names", "cprinta", "cprintz"]

import ast
from typing import Optional, Any


class ArgIdentifier(ast.NodeVisitor):
    """description"""

    def __init__(
        self,
        *args,
        verbose: bool = False,
        max_num_intermediate_unnamed_elements: int = 1,
    ):
        if len(args) < 1:
            raise ValueError("Supply at least one target function")
        self.result = {arg: {} for arg in args}
        self.verbose = verbose
        assert max_num_intermediate_unnamed_elements >= 0
        self.num_unnamed_sequence_elements = max_num_intermediate_unnamed_elements

    def visit_Call(self, node: ast.AST) -> None:
        """
        Should work for most use cases, but no guarantee

        :param node:
        :return:
        """
        if hasattr(node.func, "id") and node.func.id in self.result:
            for arg in node.args:
                if isinstance(arg, ast.Name):
                    n = arg.id
                elif isinstance(arg, ast.Call):
                    if isinstance(arg.func, ast.Attribute):
                        # print(first_arg.func.value.value.id) # TODO: UNROLLING is possible, do some resursion
                        # print(first_arg.func.value.attr) # TODO: SAME for full qualification in scope
                        n = f"{arg.func.attr}"
                    else:
                        n = f"{arg.func.id}"
                    if self.verbose:
                        args_repr = f'{", ".join([ast.dump(sub) for sub in arg.args])}'
                        kws_repr = f'{", ".join([ast.dump(sub) for sub in arg.keywords])}'
                        args_kw_repr = []
                        if len(args_repr) > 1:
                            args_kw_repr.append(args_repr)
                        if len(kws_repr) > 1:
                            args_kw_repr.append(kws_repr)
                        n += f'({", ".join(args_kw_repr)})'
                elif isinstance(arg, (ast.List, ast.Set, ast.Tuple)):
                    elts = arg.elts
                    if self.num_unnamed_sequence_elements < len(elts) - 2:
                        if (
                            self.num_unnamed_sequence_elements
                        ):  # TODO: Generalise to another external function, "pick num from sequence" func
                            stride = (len(elts) // self.num_unnamed_sequence_elements) + 1
                            between = elts[1:-2:stride]
                        else:
                            between = []
                        elts_str = [ast.dump(sub) for sub in [elts[0]] + between + [elts[-1]]]
                        n = f'[{" .. ".join(elts_str)}]'
                    else:
                        n = f'[{", ".join([ast.dump(sub) for sub in elts])}]'
                elif isinstance(arg, ast.Dict):
                    kw_repr = f'{", ".join([f"{k}:{v}" for k, v in zip([ast.dump(sub) for sub in arg.keys], [ast.dump(sub) for sub in arg.values])])}'
                    n = f"{{{kw_repr}}}"
                else:  # No obvious name
                    if self.verbose:
                        print(type(arg))
                        n = f"{ast.dump(arg)}"
                    else:
                        n = "iterable"
                if node.lineno not in self.result[node.func.id]:
                    self.result[node.func.id][node.lineno] = []
                self.result[node.func.id][node.lineno].append(n)
        self.generic_visit(node)  # visit the children


def get_arg_names(func_name: str, *, verbose=False, max_num_intermediate_unnamed_elements=1) -> Optional[str]:
    """description"""
    import inspect
    import textwrap
    import ast

    caller_frame = inspect.currentframe().f_back.f_back
    try:
        caller_src_code_lines = inspect.getsourcelines(caller_frame)
        fai = ArgIdentifier(
            func_name,
            verbose=verbose,
            max_num_intermediate_unnamed_elements=max_num_intermediate_unnamed_elements,
        )
        fai.visit(ast.parse(textwrap.dedent("".join(caller_src_code_lines[0]))))
        if func_name in fai.result:
            offset = 0
            if caller_src_code_lines[1]:
                offset = caller_src_code_lines[1] - 1
            idx = caller_frame.f_lineno - offset
            if idx in fai.result[func_name]:
                return fai.result[func_name][idx]
            elif verbose:
                print(
                    f'Unexpected line number: {idx}, probably a wrong alias "{func_name}" was supplied, found {fai.result[func_name]}, in {inspect.getsourcefile(caller_frame)}'
                )
        elif verbose:
            print(f"{func_name} was not found in {fai.result}")
    except Exception as e:
        print(e)
    return


def ge_arg_names_recurse() -> Optional[str]:
    """
    unpack chained generators to base iterator name

    :return:
    :rtype:
    """
    pass  # TODO: For e.g. description in progress_bar(range(_name_,_name2_))
    raise NotImplementedError


def cprinta(*v: Any, writer: callable = print, deliminator: str = ":") -> None:
    """

    :param v:
    :type v:
    :param writer:
    :type writer:
    :param deliminator:
    :type deliminator:
    """
    if isinstance(v, str) and v.strip() == "":
        v = '""'
    writer(f"{get_arg_names('cprinta')}{deliminator}", v)


def cprintz(*v: Any, writer: callable = print) -> None:
    """

    :param v:
    :type v:
    :param writer:
    :type writer:
    """
    if isinstance(v, str) and v.strip() == "":
        v = '""'

    gen = zip(get_arg_names("cprintz"), v)
    writer(f"{list(gen)}")


if __name__ == "__main__":

    def siajd():
        """description"""
        s = ""
        aisjd = s
        sioj = 4
        cprinta(s, aisjd, sioj)
        cprinta("")
        ass = "    "
        ls = ass + s
        cprinta(ass, ls, 2)
        cprinta("  ")

        cprintz(ass, ls, 2)

    siajd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/01/2020
           """

__all__ = [
    "kws_sink",
    "sink",
    "prod",
    "collate_first_dim",
    "call_identity",
    "args_sink",
    "identity",
    "invert_shallow_mapping",
    "flip_two_level_mapping",
    "swap_mapping_order",
    "nop",
    "empty_str",
    "list_keys",
    "first_key",
    "last_key",
]

import operator
from collections import defaultdict
from copy import deepcopy
from functools import reduce
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    Tuple,
    List,
)

from warg import Number, drop_unused_kws


def list_keys(d: Dict) -> List[Any]:
    return list(d.keys())


def first_key(d: Dict) -> Any:
    return list_keys(d)[0]


def last_key(d: Dict) -> Any:
    return list_keys(d)[-1]


def empty_str() -> str:
    """

    :return:
    :rtype:
    """
    return ""


def nop() -> None:
    """
    :rtype: None
    """
    pass


def identity(a: Any) -> Any:
    """description"""
    return a


@drop_unused_kws
def kws_sink(*args) -> Tuple[Any, ...]:
    """
    Returns args without any modification what so ever. Drops kws
    :return:"""
    return args


def call_identity(*args, **kwargs) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """

    :param args:
    :param kwargs:
    :return:"""
    return args, kwargs


# noinspection PyUnusedLocal
def args_sink(*args, **kwargs) -> Dict[str, Any]:
    """

    :param args:
    :param kwargs:
    :return:"""
    return kwargs


# noinspection PyUnusedLocal
def sink(*args, **kwargs) -> None:
    """
    Returns None, but accepts everything

    :param args:
    :param kwargs:
    :return:"""
    return


def prod(iterable: Iterable[Number]) -> Number:
    """
    Calculate the product of the a Iterable of int or floats
    :param iterable:
    :return:"""
    return reduce(operator.mul, iterable, 1)


def collate_first_dim(batch: Iterable) -> Tuple:
    """

    :param batch:
    :return:"""
    return tuple(zip(*batch))


def invert_shallow_mapping(m: Mapping) -> Dict:
    """

    :param m:
    :return:
    """
    return {v: k for k, v in m.items()}


def flip_two_level_mapping(m: Mapping) -> Dict:
    """
    result = {}
    [result.setdefault(a, {}).update({k:b}) for k, v in m.items() for a, b in v.items()]
    return result

    :param m:
    :return:
    """

    flipped = defaultdict(dict)
    for key, val in m.items():
        for sub_key, sub_val in val.items():
            flipped[sub_key][key] = sub_val
    return flipped


def swap_mapping_order(m: Mapping, order: Sequence[int]) -> Mapping:
    """

    :param m:
    :param order:
    :return:
    """
    order = [*order]

    def deep_swap(dict_, level):
        """

        :param dict_:
        :param level:
        :return:
        """

        def swap_two_level_dict(a):
            """

            :param a:
            :return:
            """
            b = defaultdict(dict)
            for key1, value1 in a.items():
                for key2, value2 in value1.items():
                    b[key2].update({key1: value2})
            return b

        dict_ = deepcopy(dict_)
        if level == 0:
            dict_ = swap_two_level_dict(dict_)
        else:
            for key in dict_:
                dict_[key] = deep_swap(dict_[key], level - 1)
        return dict_

    for pas_no in range(len(order) - 1, 0, -1):
        for i in range(pas_no):
            if order[i] > order[i + 1]:
                temp = order[i]
                order[i] = order[i + 1]
                order[i + 1] = temp
                m = deep_swap(m, i)
    return m


def chain_filter(it: Iterable, *filters: Callable) -> Iterator:
    """
    Apply a sequence of callables to an iterable through filter; filtering the iterable to the subset of a callable
    returns

    Args:
        it (Iterable):
            iterable to be filtered
        filters (Callable):
            The filter callables

    Returns:
        Iterator:
            returns an iterator yielding those items of iterable for which all(filters(item)) is true. If filters are
            None, return the items that are true.
    """
    for f in filters:
        it = filter(f, it)
    return it


def chain_apply(it: Iterable, *callables: Callable) -> Iterable:
    """
    Apply a sequence of callables to an iterable; apply the iterable sequentially in callables order

    Args:
        it (Iterable):
            iterable to be applied to
        callables (Callable):
             The applying callables

    Returns:
        Iterable:
            returns the iterable with all the callables applied.
    """
    for f in callables:
        it = f(it)
    return it


if __name__ == "__main__":

    def asud() -> None:
        """
        :rtype: None
        """
        a = {"b": 1, "h": 2}
        print(invert_shallow_mapping(a))

    def asjdnasid() -> None:
        """
        :rtype: None
        """
        a = {
            "b": {"c": {"d": [0, 1, 2], "e": [3, 4, 5, 6]}, "f": [7, 8], "g": [9]},
            "h": {"j": [10, 11]},
        }
        print(flip_two_level_mapping(a))

    def asidj() -> None:
        """
        :rtype: None
        """
        test_dict = {
            "a": {"c": {"e": 0, "f": 1}, "d": {"e": 2, "f": 3}},
            "b": {"c": {"g": 4, "h": 5}, "d": {"j": 6, "k": 7}},
        }
        result = swap_mapping_order(test_dict, [2, 0, 1])
        print(result)

    # asud()
    asidj()
    # asjdnasid()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = ""

__all__ = [
    "OrdinalIndexingDictMixin",
]


class OrdinalIndexingDictMixin:
    """
    Mixin class for indexing a class instance __dict__ (SortedDict) with both integer (ordinal) indexing or
    key:str attributes (non-ordinal) access."""

    def __getitem__(self, item):
        if isinstance(item, int):
            return list(self.__dict__.values())[item]
        else:
            return self.__dict__[item]


if __name__ == "__main__":

    def asd() -> None:
        """
        :rtype: None
        """

        class IDTM(OrdinalIndexingDictMixin):
            pass

        a = IDTM()
        a.a = 2
        a.b = 3
        assert a[0] == 2
        assert a[1] == 3



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import functools
import inspect
import types
from functools import wraps
from logging import warning
from typing import Dict, MutableMapping, Sequence, Tuple, Any

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          The concept "kw passing" implemented here lets one make a contract with the caller that all
          kwargs with be passed onwards to a receiver, this lets the caller inspect available kwargs of the
          the receiver function allowing for autocompletion, typing and documentation fetching.
           """

__all__ = [
    "passes_kws_to",
    "super_init_pass_on_kws",
    "drop_unused_kws",
    "drop_unused_args",
    "drop_kws",
    "drop_args",
    "drop_args_and_kws",
    "pack_args",
    "pack_kws",
    "pack_args_and_kws",
    "AlsoDecorator",
]


# noinspection PyUnresolvedReferences
def to_keyword_only(val: inspect.Parameter) -> inspect.Parameter:
    """

    :param val:
    :type val:
    :return:
    :rtype:"""
    if val._kind == inspect._ParameterKind.POSITIONAL_OR_KEYWORD:
        val._kind = inspect._ParameterKind.KEYWORD_ONLY
    return val


# noinspection PyUnresolvedReferences
def eval_sig_kw_params(
    passing_sig: inspect.Signature,
    receiver_func: callable,
    keep_from_var_kw: bool = False,
) -> Tuple[inspect.Signature, Dict[str, inspect.Parameter]]:
    """

    :param passing_sig:
    :type passing_sig:
    :param receiver_func:
    :type receiver_func:
    :param keep_from_var_kw:
    :type keep_from_var_kw:
    :return:
    :rtype:"""
    passing_params: dict = dict(passing_sig.parameters)
    # if inspect.isfunction()
    # if inspect.ismethod()
    # arg_spec_sig = inspect.getfullargspec(receiver_func)[0]

    receiver_params = inspect.signature(
        receiver_func
    ).parameters  # TODO: Sometime no signature is found resulting in
    # a ValueError exception

    var_kw_key = None
    var_kw = None
    for k, v in passing_params.items():
        if v.kind == inspect._ParameterKind.VAR_KEYWORD:
            var_kw_key = k

    if var_kw_key in passing_params:
        var_kw = passing_params.pop(var_kw_key)

    to_params = {
        key: to_keyword_only(val)
        for key, val in receiver_params.items()
        if val.default != inspect.Parameter.empty and key not in passing_params
    }

    passing_params.update(to_params)

    if keep_from_var_kw:
        passing_params[var_kw_key] = var_kw

        no_var_kw = True
        for k, v in receiver_params.items():
            if v.kind == inspect._ParameterKind.VAR_KEYWORD:
                no_var_kw = False
        if no_var_kw:
            warning(
                f"Receiver {receiver_func} with {receiver_params} does not acceptable arbitrary kwargs although "
                f"from_func will pass "
                f"all "
                f"kwargs onwards TypeErrors might occur, to fix this let this receiver accept any arbitrary "
                f"kwargs by adding ..,**kwargs): to the receivers function declaration"
            )

    return passing_sig, passing_params


def passes_kws_to(*receiver_funcs: callable, keep_from_var_kw: bool = False) -> callable:
    """
    A contract decorator, attaching this to a function you explicitly state that kws will be passed onward to
    a receiver function. No call graph checks if this actually enforces this yet. Also all receiver kwargs
    must be able to be received by receivers if multiple contracts are use

    :param receiver_funcs:
    :param keep_from_var_kw:
    :return:"""
    for receiver_func in receiver_funcs:
        if isinstance(receiver_func, types.BuiltinFunctionType):
            raise AssertionError(f"'Built In Receiver' function: {receiver_func}, is not supported")

    def _func(passing_func: callable) -> callable:
        passing_sig = inspect.signature(passing_func)
        for rf in receiver_funcs:
            passing_sig, new_params = eval_sig_kw_params(passing_sig, rf, keep_from_var_kw)
            passing_sig = passing_sig.replace(parameters=list(new_params.values()))
        passing_func.__signature__ = passing_sig
        return passing_func

    return _func


def super_init_pass_on_kws(
    f: callable = None, *, super_base: type = None, keep_from_var_kw: bool = False
) -> callable:
    """

    :param f:
    :param super_base:
    :param keep_from_var_kw:
    :return:"""

    def _func(func) -> callable:
        if super_base:
            to_func = super_base.__init__
        else:
            to_func = inspect.getmro(func)[0].__init__

        from_func = func.__init__

        from_sig = inspect.signature(from_func)
        from_signature, signature_parameters = eval_sig_kw_params(from_sig, to_func, keep_from_var_kw)
        from_func.__signature__ = from_signature.replace(parameters=list(signature_parameters.values()))
        return func

    if f:
        return _func(f)

    return _func


def drop_args(f: callable) -> callable:
    """

    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(f)
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        return f(**kwargs)

    return wrapper


def drop_kws(f: callable) -> callable:
    """

    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(f)
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        return f(*args)

    return wrapper


def drop_args_and_kws(f: callable) -> callable:
    """

    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(f)
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        return f()

    return wrapper


WRAPPER_NO_ANNOTATION = tuple(
    set(functools.WRAPPER_ASSIGNMENTS)
    - {
        "__annotations__",
    }
)


def pack_args(
    f: callable,
    *,
    pack_name: str = "arg_pack",
    allow_passing: bool = True,
    verbose: bool = False,
) -> callable:
    """

    :param pack_name:
    :type pack_name:
    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(
        f,
        # assigned=WRAPPER_NO_ANNOTATION,
        # updated=("__annotations__",),
    )
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        new_kwargs = kwargs.copy()
        if not allow_passing:
            assert pack_name not in kwargs, f"thou shall not pass {pack_name}"
        else:
            if pack_name in kwargs:
                if verbose:
                    print(f"{pack_name} was extended, careful!")
                a = kwargs.pop(pack_name, None)
                new_kwargs[pack_name] = (*a, *args)
            else:
                new_kwargs[pack_name] = args
        return f(*args, **new_kwargs)

    return wrapper


def pack_kws(
    f: callable,
    *,
    pack_name: str = "kw_pack",
    allow_passing: bool = True,
    verbose: bool = False,
) -> callable:
    """

    :param pack_name:
    :type pack_name:
    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(
        f,
        # assigned=WRAPPER_NO_ANNOTATION, #TODO: Figure out if pack_name can be hidden from function signature
        # updated=("__annotations__",),
    )
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        new_kwargs = kwargs.copy()
        if not allow_passing:
            assert pack_name not in kwargs, f"thou shall not pass {pack_name}"
        else:  # TODO: else keyword can be removed, but branch remain
            if pack_name in kwargs:
                if verbose:
                    print(f"{pack_name} was extended, careful!")
                k = kwargs.pop(pack_name, None)
                new_kwargs[pack_name] = {**k, **kwargs}
            else:
                new_kwargs[pack_name] = kwargs
        return f(*args, **new_kwargs)

    return wrapper


def pack_args_and_kws(
    f: callable,
    *,
    pack_name: str = "arg_kw_pack",
    allow_passing: bool = True,
    verbose: bool = False,
) -> callable:
    """

    :param pack_name:
    :type pack_name:
    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(
        f,
        # assigned=WRAPPER_NO_ANNOTATION,
        # updated=("__annotations__",),
    )
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        new_kwargs = kwargs.copy()
        if not allow_passing:
            assert pack_name not in kwargs, f"thou shall not pass {pack_name}"
        else:  # TODO: else keyword can be removed, but branch remain
            if pack_name in kwargs:
                if verbose:
                    print(f"{pack_name} was extended, careful!")
                a, k = kwargs.pop(pack_name, None)
                new_kwargs[pack_name] = ((*a, *args), {**k, **kwargs})
            else:
                new_kwargs[pack_name] = (args, kwargs)
        return f(*args, **kwargs)

    return wrapper


def drop_unused_args(f: callable) -> callable:
    """

    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(f)
    def wrapper(*args, **kwargs: MutableMapping) -> Any:
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        return f(**kwargs)

    return wrapper


# noinspection PyUnresolvedReferences
def drop_unused_kws(f: callable) -> callable:
    """

    :param f:
    :type f:
    :return:
    :rtype:"""

    @wraps(f)
    def wrapper(*args, **kwargs: MutableMapping):
        """

        :param args:
        :type args:
        :return:
        :rtype:"""
        from_sig = inspect.signature(f)

        for k, v in from_sig.parameters.items():
            if v.kind == inspect._ParameterKind.VAR_KEYWORD:
                return f(*args, **kwargs)

        kept = {}
        for k, v in kwargs.items():
            if k in from_sig.parameters.keys():
                kept[k] = v

        return f(*args, **kept)

    return wrapper


class AlsoDecorator:
    """
    Lets you use a function as a decorator too
    """

    def __call__(self, func):
        @functools.wraps(func)
        def decorate_func(*args: Sequence, **kwargs: MutableMapping):
            """

            :param args:
            :type args:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:"""
            with self:
                return func(*args, **kwargs)

        return decorate_func


if __name__ == "__main__":

    def _main():
        class BaseClass:
            """description"""

            def __init__(self, arg0, *args, kwarg0=None, kwarg1=None, **kwargs: MutableMapping):
                self.arg0 = arg0
                for key, val in enumerate(args):
                    setattr(self, f"arg{key + 1}", val)
                self.kwarg0 = kwarg0
                self.kwarg1 = kwarg1
                self.__dict__.update(kwargs)

        class SubClass0(BaseClass):
            """description"""

            @passes_kws_to(BaseClass.__init__)
            def __init__(self, arg0, arg1, arg2, *args, kwarg2=None, **kwargs: MutableMapping):
                super().__init__(arg0, *args, **kwargs)
                self.arg1 = arg1
                self.arg2 = arg2
                self.kwarg2 = kwarg2

        @super_init_pass_on_kws
        class SubClass1(BaseClass):
            """description"""

            def __init__(self, arg0, arg1, arg2, *args, kwarg2=None, **kwargs: MutableMapping):
                super().__init__(arg0, *args, **kwargs)
                self.arg1 = arg1
                self.arg2 = arg2
                self.kwarg2 = kwarg2

        @super_init_pass_on_kws(super_base=BaseClass)
        class SubClass2(BaseClass):
            """description"""

            def __init__(self, arg0, arg1, arg2, *args, kwarg2=None, **kwargs: MutableMapping):
                super().__init__(arg0, *args, **kwargs)
                self.arg1 = arg1
                self.arg2 = arg2
                self.kwarg2 = kwarg2

        @drop_unused_kws
        def some_func(*, a):
            """

            :param a:
            :type a:"""
            print(a)

        @drop_unused_kws
        def some_other_func(*, a, **kwargs: MutableMapping):
            """

            :param a:
            :type a:
            :param kwargs:
            :type kwargs:"""
            print(a, kwargs)

        @drop_unused_kws
        def some_different_func(*, a, b):
            """

            :param a:
            :type a:
            :param b:
            :type b:"""
            print(a, b)

        print(inspect.signature(SubClass0.__init__))
        print(inspect.signature(SubClass1.__init__))
        print(inspect.signature(SubClass1.__init__))

        print(vars(SubClass0(1, 1, 1, kwarg0=52)))
        print(vars(SubClass1(2, 2, 1, kwarg0=52)))
        print(vars(SubClass2(1, 1, 1, kwarg0=52)))
        print(inspect.getmro(SubClass0))

        some_func(a=1, b=2, c=3)

        some_other_func(a=1, b=2)

        some_different_func(a=1, c=2, b="l")

    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04-01-2021
           """

import enum
import sys

__all__ = [
    "get_backend_module",
    "is_py3",
    "get_system",
    # "set_system",
    "SystemEnum",
]

from types import ModuleType

if sys.platform.startswith("java"):
    import platform

    os_name = platform.java_ver()[3][0]
    if os_name.startswith("Windows"):  # "Windows XP", "Windows 7", etc.
        SYSTEM_ = "win32"
    elif os_name.startswith("Mac"):  # "Mac OS X", etc.
        SYSTEM_ = "darwin"
    else:  # "Linux", "SunOS", "FreeBSD", etc.
        # Setting this to "linux2" is not ideal, but only Windows or Mac
        # are actually checked for and the rest of the module expects
        # *sys.platform* style strings.
        SYSTEM_ = "linux2"
else:
    SYSTEM_ = sys.platform

SYSTEM = SYSTEM_


class SystemEnum(enum.Enum):
    """
    Enum for the system type
    """

    windows, linux, mac, other = range(4)


def set_system(system: SystemEnum) -> None:
    """

    :param system:
    :type system:
    """
    global SYSTEM
    SYSTEM = system


def get_system() -> SystemEnum:
    """

    :return:
    :rtype:
    """
    return SYSTEM


if SYSTEM == "darwin":
    SYSTEM = SystemEnum.mac
elif SYSTEM == "linux2" or SYSTEM == "linux":
    SYSTEM = SystemEnum.linux
elif SYSTEM == "win32":
    SYSTEM = SystemEnum.windows
else:
    SYSTEM = SystemEnum.other


def get_backend_module(project_name: str, backend_name: str = sys.platform) -> ModuleType:
    """Returns the backend module.

    :param project_name:
    :type project_name:
    :param backend_name:
    :type backend_name:
    :return:
    :rtype:

    """
    import importlib

    try:
        importlib.import_module(f"{project_name}")
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"{project_name} not found, please install it")

    modules = []
    if backend_name is not None:
        modules += [backend_name]
    elif sys.platform == "darwin":
        modules += ["darwin"]
    elif sys.platform == "win32":
        modules += ["win10"]
    else:
        modules += [
            "appindicator",
            "gtk",
            "xorg",
            "gtk_dbus"
            # "unity", "kde", "gnome", "fallback",
        ]

    errors = []
    for module in modules:
        try:
            return importlib.import_module(f"{project_name}.{module}")
        except ImportError as e:
            errors.append(e)

    # Did not find any backend, raise error
    raise ImportError(f'{sys.platform} platform is not supported: {"; ".join(str(e) for e in errors)}')


def is_py3():
    """

    :return:
    :rtype:
    """
    return sys.version_info[0] == 3


if __name__ == "__main__":
    print(get_backend_module("draugr", "python_utilities"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
Modified for all import

An OrderedSet is a custom MutableSet that remembers its order, so that every
entry has an index that can be looked up. It can also act like a Sequence.

Based on a recipe originally posted to ActiveState Recipes by Raymond Hettiger,
and released under the MIT license.

           Created on 06-11-2020
           """

__all__ = ["OrderedSet"]

from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    MutableSet,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Union,
    overload,
)

SLICE_ALL = slice(None)
__version__ = "4.0.2"

T = TypeVar("T")


# When Python 3.6 is the minimum version, we can define a type like this,
# parameterizing the types that an OrderedSet can interoperate with:
#
# SetLike = Union[Sequence[T], Set[T]]


def _is_atomic(obj: Any) -> bool:
    """
    Returns True for objects which are iterable but should not be iterated in
    the context of indexing an OrderedSet.

    When we index by an iterable, usually that means we're being asked to look
    up a list of things.

    However, in the case of the .index() method, we shouldn't handle strings
    and tuples like other iterables. They're not sequences of things to look
    up, they're the single, atomic thing we're trying to find.

    As an example, oset.index('hello') should give the index of 'hello' in an
    OrderedSet of strings. It shouldn't give the indexes of each individual
    character.
    """
    return isinstance(obj, str) or isinstance(obj, tuple)


class OrderedSet(MutableSet[T], Sequence[T]):
    """
    An OrderedSet is a custom MutableSet that remembers its order, so that
    every entry has an index that can be looked up.

    Example:
        >>> OrderedSet([1, 1, 2, 3, 2])
        OrderedSet([1, 2, 3])
    """

    def __init__(self, iterable: Optional[Iterable[T]] = None):
        self.items = []  # type: List[T]
        self.map = {}  # type: Dict[T, int]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        """
        Returns the number of unique elements in the ordered set

        Example:
            >>> len(OrderedSet([]))
            0
            >>> len(OrderedSet([1, 2]))
            2
        """
        return len(self.items)

    @overload
    def __getitem__(self, index: Sequence[int]) -> List[T]:
        ...

    @overload
    def __getitem__(self, index: slice) -> "OrderedSet[T]":
        ...

    def __getitem__(self, index: int) -> T:
        """
        Get the item at a given index.

        If `index` is a slice, you will get back that slice of items, as a
        new OrderedSet.

        If `index` is a list or a similar iterable, you'll get a list of
        items corresponding to those indices. This is similar to NumPy's
        "fancy indexing". The result is not an OrderedSet because you may ask
        for duplicate indices, and the number of elements returned should be
        the number of elements asked for.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset[1]
            2
        """
        if isinstance(index, slice) and index == SLICE_ALL:
            return self.copy()
        elif isinstance(index, Iterable):
            return [self.items[i] for i in index]
        elif isinstance(index, slice) or hasattr(index, "__index__"):
            result = self.items[index]
            if isinstance(result, list):
                return self.__class__(result)
            else:
                return result
        else:
            raise TypeError("Don't know how to index an OrderedSet by %r" % index)

    def copy(self) -> "OrderedSet[T]":
        """
        Return a shallow copy of this object.

        Example:
            >>> this = OrderedSet([1, 2, 3])
            >>> other = this.copy()
            >>> this == other
            True
            >>> this is other
            False
        """
        return self.__class__(self)

    # Define the gritty details of how an OrderedSet is serialized as a pickle.
    # We leave off type annotations, because the only code that should interact
    # with these is a generalized tool such as pickle.
    def __getstate__(self):
        if len(self) == 0:
            # In pickle, the state can't be an empty list.
            # We need to return a truthy value, or else __setstate__ won't be run.
            #
            # This could have been done more gracefully by always putting the state
            # in a tuple, but this way is backwards- and forwards- compatible with
            # previous versions of OrderedSet.
            return (None,)
        else:
            return list(self)

    def __setstate__(self, state):
        if state == (None,):
            self.__init__([])
        else:
            self.__init__(state)

    def __contains__(self, key: Any) -> bool:
        """
        Test if the item is in this ordered set.

        Example:
            >>> 1 in OrderedSet([1, 3, 2])
            True
            >>> 5 in OrderedSet([1, 3, 2])
            False
        """
        return key in self.map

    def add(self, key: T) -> int:
        """
        Add `key` as an item to this OrderedSet, then return its index.

        If `key` is already in the OrderedSet, return the index it already
        had.

        Example:
            >>> oset = OrderedSet()
            >>> oset.append(3)
            0
            >>> print(oset)
            OrderedSet([3])
        """
        if key not in self.map:
            self.map[key] = len(self.items)
            self.items.append(key)
        return self.map[key]

    append = add

    def update(self, sequence: Union[Sequence[T], Set[T]]) -> int:
        """
        Update the set with the given iterable sequence, then return the index
        of the last element inserted.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.update([3, 1, 5, 1, 4])
            4
            >>> print(oset)
            OrderedSet([1, 2, 3, 5, 4])
        """
        item_index = 0
        try:
            for item in sequence:
                item_index = self.add(item)
        except TypeError:
            raise ValueError(f"Argument needs to be an iterable, got {type(sequence)}")
        return item_index

    @overload
    def index(self, key: T) -> int:
        """description"""
        ...

    def index(self, key: Sequence[T]) -> int:
        """
        Get the index of a given entry, raising an IndexError if it's not
        present.

        `key` can be an iterable of entries that is not a string, in which case
        this returns a list of indices.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.index(2)
            1
        """
        if isinstance(key, Iterable) and not _is_atomic(key):
            return [self.index(subkey) for subkey in key]
        return self.map[key]

    # Provide some compatibility with pandas.Index
    get_loc = index
    get_indexer = index

    def pop(self) -> T:
        """
        Remove and return the last element from the set.

        Raises KeyError if the set is empty.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.pop()
            3
        """
        if not self.items:
            raise KeyError("Set is empty")

        elem = self.items[-1]
        del self.items[-1]
        del self.map[elem]
        return elem

    def discard(self, key: T) -> None:
        """
        Remove an element.  Do not raise an exception if absent.

        The MutableSet mixin uses this to implement the .remove() method, which
        *does* raise an error when asked to remove a non-existent item.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.discard(2)
            >>> print(oset)
            OrderedSet([1, 3])
            >>> oset.discard(2)
            >>> print(oset)
            OrderedSet([1, 3])
        """
        if key in self:
            i = self.map[key]
            del self.items[i]
            del self.map[key]
            for k, v in self.map.items():
                if v >= i:
                    self.map[k] = v - 1

    def clear(self) -> None:
        """
        Remove all items from this OrderedSet.
        """
        del self.items[:]
        self.map.clear()

    def __iter__(self) -> Iterator[T]:
        """
        Example:
            >>> list(iter(OrderedSet([1, 2, 3])))
            [1, 2, 3]
        """
        return iter(self.items)

    def __reversed__(self) -> Iterator[T]:
        """
        Example:
            >>> list(reversed(OrderedSet([1, 2, 3])))
            [3, 2, 1]
        """
        return reversed(self.items)

    def __repr__(self) -> str:
        if not self:
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}({list(self)!r})"

    def __eq__(self, other: Any) -> bool:
        """
        Returns true if the containers have the same items. If `other` is a
        Sequence, then order is checked, otherwise it is ignored.

        Example:
            >>> oset = OrderedSet([1, 3, 2])
            >>> oset == [1, 3, 2]
            True
            >>> oset == [1, 2, 3]
            False
            >>> oset == [2, 3]
            False
            >>> oset == OrderedSet([3, 2, 1])
            False
        """
        if isinstance(other, Sequence):
            # Check that this OrderedSet contains the same elements, in the
            # same order, as the other object.
            return list(self) == list(other)
        try:
            other_as_set = set(other)
        except TypeError:
            # If `other` can't be converted into a set, it's not equal.
            return False
        else:
            return set(self) == other_as_set

    def union(self, *sets: Union[Sequence[T], Set[T]]) -> "OrderedSet[T]":
        """
        Combines all unique items.
        Each items order is defined by its first appearance.

        Example:
            >>> oset = OrderedSet.union(OrderedSet([3, 1, 4, 1, 5]), [1, 3], [2, 0])
            >>> print(oset)
            OrderedSet([3, 1, 4, 5, 2, 0])
            >>> oset.union([8, 9])
            OrderedSet([3, 1, 4, 5, 2, 0, 8, 9])
            >>> oset | {10}
            OrderedSet([3, 1, 4, 5, 2, 0, 10])
        """
        cls = self.__class__ if isinstance(self, OrderedSet) else OrderedSet
        containers = map(list, it.chain([self], sets))
        items = it.chain.from_iterable(containers)
        return cls(items)

    def __and__(self, other: Union[Sequence[T], Set[T]]) -> "OrderedSet[T]":
        # the parent implementation of this is backwards
        return self.intersection(other)

    def intersection(self, *sets: Union[Sequence[T], Set[T]]) -> "OrderedSet[T]":
        """
        Returns elements in common between all sets. Order is defined only
        by the first set.

        Example:
            >>> oset = OrderedSet.intersection(OrderedSet([0, 1, 2, 3]), [1, 2, 3])
            >>> print(oset)
            OrderedSet([1, 2, 3])
            >>> oset.intersection([2, 4, 5], [1, 2, 3, 4])
            OrderedSet([2])
            >>> oset.intersection()
            OrderedSet([1, 2, 3])
        """
        cls = self.__class__ if isinstance(self, OrderedSet) else OrderedSet
        if sets:
            common = set.intersection(*map(set, sets))
            items = (item for item in self if item in common)
        else:
            items = self
        return cls(items)

    def difference(self, *sets: Union[Sequence[T], Set[T]]) -> "OrderedSet[T]":
        """
        Returns all elements that are in this set but not the others.

        Example:
            >>> OrderedSet([1, 2, 3]).difference(OrderedSet([2]))
            OrderedSet([1, 3])
            >>> OrderedSet([1, 2, 3]).difference(OrderedSet([2]), OrderedSet([3]))
            OrderedSet([1])
            >>> OrderedSet([1, 2, 3]) - OrderedSet([2])
            OrderedSet([1, 3])
            >>> OrderedSet([1, 2, 3]).difference()
            OrderedSet([1, 2, 3])
        """
        cls = self.__class__
        if sets:
            other = set.union(*map(set, sets))
            items = (item for item in self if item not in other)
        else:
            items = self
        return cls(items)

    def issubset(self, other: Union[Sequence[T], Set[T]]) -> bool:
        """
        Report whether another set contains this set.

        Example:
            >>> OrderedSet([1, 2, 3]).issubset({1, 2})
            False
            >>> OrderedSet([1, 2, 3]).issubset({1, 2, 3, 4})
            True
            >>> OrderedSet([1, 2, 3]).issubset({1, 4, 3, 5})
            False
        """
        if len(self) > len(other):  # Fast check for obvious cases
            return False
        return all(item in other for item in self)

    def issuperset(self, other: Union[Sequence[T], Set[T]]) -> bool:
        """
        Report whether this set contains another set.

        Example:
            >>> OrderedSet([1, 2]).issuperset([1, 2, 3])
            False
            >>> OrderedSet([1, 2, 3, 4]).issuperset({1, 2, 3})
            True
            >>> OrderedSet([1, 4, 3, 5]).issuperset({1, 2, 3})
            False
        """
        if len(self) < len(other):  # Fast check for obvious cases
            return False
        return all(item in self for item in other)

    def symmetric_difference(self, other: Union[Sequence[T], Set[T]]) -> "OrderedSet[T]":
        """
        Return the symmetric difference of two OrderedSets as a new set.
        That is, the new set will contain all elements that are in exactly
        one of the sets.

        Their order will be preserved, with elements from `self` preceding
        elements from `other`.

        Example:
            >>> this = OrderedSet([1, 4, 3, 5, 7])
            >>> other = OrderedSet([9, 7, 1, 3, 2])
            >>> this.symmetric_difference(other)
            OrderedSet([4, 5, 9, 2])
        """
        cls = self.__class__ if isinstance(self, OrderedSet) else OrderedSet
        diff1 = cls(self).difference(other)
        diff2 = cls(other).difference(self)
        return diff1.union(diff2)

    def _update_items(self, items: list) -> None:
        """
        Replace the 'items' list of this OrderedSet with a new one, updating
        self.map accordingly.
        """
        self.items = items
        self.map = {item: idx for (idx, item) in enumerate(items)}

    def difference_update(self, *sets: Union[Sequence[T], Set[T]]) -> None:
        """
        Update this OrderedSet to remove items from one or more other sets.

        Example:
            >>> this = OrderedSet([1, 2, 3])
            >>> this.difference_update(OrderedSet([2, 4]))
            >>> print(this)
            OrderedSet([1, 3])

            >>> this = OrderedSet([1, 2, 3, 4, 5])
            >>> this.difference_update(OrderedSet([2, 4]), OrderedSet([1, 4, 6]))
            >>> print(this)
            OrderedSet([3, 5])
        """
        items_to_remove = set()  # type: Set[T]
        for other in sets:
            items_as_set = set(other)  # type: Set[T]
            items_to_remove |= items_as_set
        self._update_items([item for item in self.items if item not in items_to_remove])

    def intersection_update(self, other: Union[Sequence[T], Set[T]]) -> None:
        """
        Update this OrderedSet to keep only items in another set, preserving
        their order in this set.

        Example:
            >>> this = OrderedSet([1, 4, 3, 5, 7])
            >>> other = OrderedSet([9, 7, 1, 3, 2])
            >>> this.intersection_update(other)
            >>> print(this)
            OrderedSet([1, 3, 7])
        """
        other = set(other)
        self._update_items([item for item in self.items if item in other])

    def symmetric_difference_update(self, other: Union[Sequence[T], Set[T]]) -> None:
        """
        Update this OrderedSet to remove items from another set, then
        add items from the other set that were not present in this set.

        Example:
            >>> this = OrderedSet([1, 4, 3, 5, 7])
            >>> other = OrderedSet([9, 7, 1, 3, 2])
            >>> this.symmetric_difference_update(other)
            >>> print(this)
            OrderedSet([4, 5, 9, 2])
        """
        items_to_add = [item for item in other if item not in self]
        items_to_remove = set(other)
        self._update_items([item for item in self.items if item not in items_to_remove] + items_to_add)


if __name__ == "__main__":
    import collections
    import itertools as it
    import operator
    import pickle
    import random
    import sys

    import pytest

    def test_pickle():
        set1 = OrderedSet("abracadabra")
        roundtrip = pickle.loads(pickle.dumps(set1))
        assert roundtrip == set1

    def test_empty_pickle():
        empty_oset = OrderedSet()
        empty_roundtrip = pickle.loads(pickle.dumps(empty_oset))
        assert empty_roundtrip == empty_oset

    def test_order():
        set1 = OrderedSet("abracadabra")
        assert len(set1) == 5
        assert set1 == OrderedSet(["a", "b", "r", "c", "d"])
        assert list(reversed(set1)) == ["d", "c", "r", "b", "a"]

    def test_binary_operations():
        set1 = OrderedSet("abracadabra")
        set2 = OrderedSet("simsalabim")
        assert set1 != set2

        assert set1 & set2 == OrderedSet(["a", "b"])
        assert set1 | set2 == OrderedSet(["a", "b", "r", "c", "d", "s", "i", "m", "l"])
        assert set1 - set2 == OrderedSet(["r", "c", "d"])

    def test_indexing():
        set1 = OrderedSet("abracadabra")
        assert set1[:] == set1
        assert set1.copy() == set1
        assert set1 is set1
        assert set1[:] is not set1
        assert set1.copy() is not set1

        assert set1[[1, 2]] == OrderedSet(["b", "r"])
        assert set1[1:3] == OrderedSet(["b", "r"])
        assert set1.index("b") == 1
        assert set1.index(["b", "r"]) == [1, 2]
        with pytest.raises(KeyError):
            set1.index("br")

    class FancyIndexTester:
        """
        Make sure we can index by a NumPy ndarray, without having to import
        NumPy.
        """

        def __init__(self, indices):
            self.indices = indices

        def __iter__(self):
            return iter(self.indices)

        def __index__(self):
            raise TypeError("NumPy arrays have weird __index__ methods")

        def __eq__(self, other):
            # Emulate NumPy being fussy about the == operator
            raise TypeError

    def test_fancy_index_class():
        set1 = OrderedSet("abracadabra")
        indexer = FancyIndexTester([1, 0, 4, 3, 0, 2])
        assert "".join(set1[indexer]) == "badcar"

    def test_pandas_compat():
        set1 = OrderedSet("abracadabra")
        assert set1.get_loc("b") == 1
        assert set1.get_indexer(["b", "r"]) == [1, 2]

    def test_tuples():
        set1 = OrderedSet()
        tup = ("tuple", 1)
        set1.add(tup)
        assert set1.index(tup) == 0
        assert set1[0] == tup

    def test_remove():
        set1 = OrderedSet("abracadabra")

        set1.remove("a")
        set1.remove("b")

        assert set1 == OrderedSet("rcd")
        assert set1[0] == "r"
        assert set1[1] == "c"
        assert set1[2] == "d"

        assert set1.index("r") == 0
        assert set1.index("c") == 1
        assert set1.index("d") == 2

        assert "a" not in set1
        assert "b" not in set1
        assert "r" in set1

        # Make sure we can .discard() something that's already gone, plus
        # something that was never there
        set1.discard("a")
        set1.discard("a")

    def test_remove_error():
        # If we .remove() an element that's not there, we get a KeyError
        set1 = OrderedSet("abracadabra")
        with pytest.raises(KeyError):
            set1.remove("z")

    def test_clear():
        set1 = OrderedSet("abracadabra")
        set1.clear()

        assert len(set1) == 0
        assert set1 == OrderedSet()

    def test_update():
        set1 = OrderedSet("abcd")
        result = set1.update("efgh")

        assert result == 7
        assert len(set1) == 8
        assert "".join(set1) == "abcdefgh"

        set2 = OrderedSet("abcd")
        result = set2.update("cdef")
        assert result == 5
        assert len(set2) == 6
        assert "".join(set2) == "abcdef"

    def test_pop():
        set1 = OrderedSet("ab")
        elem = set1.pop()

        assert elem == "b"
        elem = set1.pop()

        assert elem == "a"

        pytest.raises(KeyError, set1.pop)

    def test_getitem_type_error():
        set1 = OrderedSet("ab")
        with pytest.raises(TypeError):
            set1["a"]

    def test_update_value_error():
        set1 = OrderedSet("ab")
        with pytest.raises(ValueError):
            # noinspection PyTypeChecker
            set1.update(3)

    def test_empty_repr():
        set1 = OrderedSet()
        assert repr(set1) == "OrderedSet()"

    def test_eq_wrong_type():
        set1 = OrderedSet()
        assert set1 != 2

    def test_ordered_equality():
        # Ordered set checks order against sequences.
        assert OrderedSet([1, 2]) == OrderedSet([1, 2])
        assert OrderedSet([1, 2]) == [1, 2]
        assert OrderedSet([1, 2]) == (1, 2)
        assert OrderedSet([1, 2]) == collections.deque([1, 2])

    def test_ordered_inequality():
        # Ordered set checks order against sequences.
        assert OrderedSet([1, 2]) != OrderedSet([2, 1])

        assert OrderedSet([1, 2]) != [2, 1]
        assert OrderedSet([1, 2]) != [2, 1, 1]

        assert OrderedSet([1, 2]) != (2, 1)
        assert OrderedSet([1, 2]) != (2, 1, 1)

        # Note: in Python 2.7 deque does not inherit from Sequence, but __eq__
        # contains an explicit check for this case for python 2/3 compatibility.
        assert OrderedSet([1, 2]) != collections.deque([2, 1])
        assert OrderedSet([1, 2]) != collections.deque([2, 2, 1])

    def test_comparisons():
        # Comparison operators on sets actually test for subset and superset.
        assert OrderedSet([1, 2]) < OrderedSet([1, 2, 3])
        assert OrderedSet([1, 2]) > OrderedSet([1])

        # MutableSet subclasses aren't comparable to set on 3.3.
        if tuple(sys.version_info) >= (3, 4):
            assert OrderedSet([1, 2]) > {1}

    def test_unordered_equality():
        # Unordered set checks order against non-sequences.
        assert OrderedSet([1, 2]) == {1, 2}
        assert OrderedSet([1, 2]) == frozenset([2, 1])

        assert OrderedSet([1, 2]) == {1: "a", 2: "b"}
        assert OrderedSet([1, 2]) == {1: 1, 2: 2}.keys()
        assert OrderedSet([1, 2]) == {1: 1, 2: 2}.values()

        # Corner case: OrderedDict is not a Sequence, so we don't check for order,
        # even though it does have the concept of order.
        assert OrderedSet([1, 2]) == collections.OrderedDict([(2, 2), (1, 1)])

        # Corner case: We have to treat iterators as unordered because there
        # is nothing to distinguish an ordered and unordered iterator
        assert OrderedSet([1, 2]) == iter([1, 2])
        assert OrderedSet([1, 2]) == iter([2, 1])
        assert OrderedSet([1, 2]) == iter([2, 1, 1])

    def test_unordered_inequality():
        assert OrderedSet([1, 2]) != set([])
        assert OrderedSet([1, 2]) != frozenset([2, 1, 3])

        assert OrderedSet([1, 2]) != {2: "b"}
        assert OrderedSet([1, 2]) != {1: 1, 4: 2}.keys()
        assert OrderedSet([1, 2]) != {1: 1, 2: 3}.values()

        # Corner case: OrderedDict is not a Sequence, so we don't check for order,
        # even though it does have the concept of order.
        assert OrderedSet([1, 2]) != collections.OrderedDict([(2, 2), (3, 1)])

    def allsame_(iterable, eq=operator.eq):
        """returns True of all items in iterable equal each other"""
        iter_ = iter(iterable)
        try:
            first = next(iter_)
        except StopIteration:
            return True
        return all(eq(first, item) for item in iter_)

    def check_results_(results, datas, name):
        """
        helper for binary operator tests.

        check that all results have the same value, but are different items.
        data and name are used to indicate what sort of tests is run.
        """
        if not allsame_(results):
            raise AssertionError("Not all same {} for {} with datas={}".format(results, name, datas))
        for a, b in it.combinations(results, 2):
            if not isinstance(a, (bool, int)):
                assert a is not b, name + " should all be different items"

    def _operator_consistency_testdata():
        """
        Predefined and random data used to test operator consistency.
        """
        # test case 1
        data1 = OrderedSet([5, 3, 1, 4])
        data2 = OrderedSet([1, 4])
        yield data1, data2

        # first set is empty
        data1 = OrderedSet([])
        data2 = OrderedSet([3, 1, 2])
        yield data1, data2

        # second set is empty
        data1 = OrderedSet([3, 1, 2])
        data2 = OrderedSet([])
        yield data1, data2

        # both sets are empty
        data1 = OrderedSet([])
        data2 = OrderedSet([])
        yield data1, data2

        # random test cases
        rng = random.Random(0)
        a, b = 20, 20
        for _ in range(10):
            data1 = OrderedSet(rng.randint(0, a) for _ in range(b))
            data2 = OrderedSet(rng.randint(0, a) for _ in range(b))
            yield data1, data2
            yield data2, data1

    def test_operator_consistency_isect():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1.copy()
            result1.intersection_update(data2)
            result2 = data1 & data2
            result3 = data1.intersection(data2)
            check_results_([result1, result2, result3], datas=(data1, data2), name="isect")

    def test_operator_consistency_difference():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1.copy()
            result1.difference_update(data2)
            result2 = data1 - data2
            result3 = data1.difference(data2)
            check_results_([result1, result2, result3], datas=(data1, data2), name="difference")

    def test_operator_consistency_xor():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1.copy()
            result1.symmetric_difference_update(data2)
            result2 = data1 ^ data2
            result3 = data1.symmetric_difference(data2)
            check_results_([result1, result2, result3], datas=(data1, data2), name="xor")

    def test_operator_consistency_union():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1.copy()
            result1.update(data2)
            result2 = data1 | data2
            result3 = data1.union(data2)
            check_results_([result1, result2, result3], datas=(data1, data2), name="union")

    def test_operator_consistency_subset():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1 <= data2
            result2 = data1.issubset(data2)
            result3 = set(data1).issubset(set(data2))
            check_results_([result1, result2, result3], datas=(data1, data2), name="subset")

    def test_operator_consistency_superset():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1 >= data2
            result2 = data1.issuperset(data2)
            result3 = set(data1).issuperset(set(data2))
            check_results_([result1, result2, result3], datas=(data1, data2), name="superset")

    def test_operator_consistency_disjoint():
        for data1, data2 in _operator_consistency_testdata():
            result1 = data1.isdisjoint(data2)
            result2 = len(data1.intersection(data2)) == 0
            check_results_([result1, result2], datas=(data1, data2), name="disjoint")

    def test_bitwise_and_consistency():
        # Specific case that was failing without explicit __and__ definition
        data1 = OrderedSet([12, 13, 1, 8, 16, 15, 9, 11, 18, 6, 4, 3, 19, 17])
        data2 = OrderedSet([19, 4, 9, 3, 2, 10, 15, 17, 11, 13, 20, 6, 14, 16, 8])
        result1 = data1.copy()
        result1.intersection_update(data2)
        # This requires a custom & operation apparently
        result2 = data1 & data2
        result3 = data1.intersection(data2)
        check_results_([result1, result2, result3], datas=(data1, data2), name="isect")
